# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import MDAnalysis as mda
import numpy as np
import sys

from scipy.io import netcdf_file

import pytest
from numpy.testing import assert_equal, assert_almost_equal
from MDAnalysis.coordinates.TRJ import NCDFReader, NCDFWriter

from MDAnalysisTests.datafiles import (
    PFncdf_Top,
    PFncdf_Trj,
    GRO,
    TRR,
    XYZ_mini,
    PRM_NCBOX,
    TRJ_NCBOX,
    DLP_CONFIG,
    CPPTRAJ_TRAJ_TOP,
    CPPTRAJ_TRAJ,
)
from MDAnalysisTests.coordinates.test_trj import _TRJReaderTest
from MDAnalysisTests.coordinates.reference import RefVGV, RefTZ2
from MDAnalysisTests import make_Universe
from MDAnalysisTests.util import block_import


class _NCDFReaderTest(_TRJReaderTest):
    prec = 3

    @pytest.fixture()
    def universe(self):
        return mda.Universe(self.topology, self.filename)

    def test_slice_iteration(self, universe):
        frames = [ts.frame for ts in universe.trajectory[4:-2:4]]
        assert_equal(
            frames,
            np.arange(universe.trajectory.n_frames)[4:-2:4],
            err_msg="slicing did not produce the expected frames",
        )

    def test_metadata(self, universe):
        data = universe.trajectory.trjfile
        assert_equal(data.Conventions.decode("utf-8"), "AMBER")
        assert_equal(data.ConventionVersion.decode("utf-8"), "1.0")

    def test_dt(self, universe):
        ref = 0.0
        assert_almost_equal(ref, universe.trajectory.dt, self.prec)
        assert_almost_equal(ref, universe.trajectory.ts.dt, self.prec)

    def test_get_writer(self, universe):
        with universe.trajectory.Writer("out.ncdf") as w:
            assert w.n_atoms == len(universe.atoms)
            assert w.remarks.startswith("AMBER NetCDF format")

    def test_get_writer_custom_n_atoms(self, universe):
        with universe.trajectory.Writer(
            "out.ncdf", n_atoms=42, remarks="Hi!"
        ) as w:
            assert w.n_atoms == 42
            assert w.remarks == "Hi!"

    def test_wrong_natoms(self):
        with pytest.raises(ValueError):
            mda.coordinates.TRJ.NCDFReader(self.filename, n_atoms=2)

    def test_read_on_closed(self, universe):
        universe.trajectory.close()

        with pytest.raises(IOError):
            universe.trajectory.__getitem__(2)

    def test_mmap_kwarg(self, universe):
        # default is None
        assert universe.trajectory._mmap == None


# Ugly way to create the tests for mmap


class _NCDFReaderTest_mmap_None(_NCDFReaderTest):
    @pytest.fixture()
    def universe(self):
        return mda.Universe(self.topology, self.filename, mmap=None)


class _NCDFReaderTest_mmap_True(_NCDFReaderTest):
    @pytest.fixture()
    def universe(self):
        return mda.Universe(self.topology, self.filename, mmap=True)

    def test_mmap_kwarg(self, universe):
        # default is None
        assert universe.trajectory._mmap == True


class _NCDFReaderTest_mmap_False(_NCDFReaderTest):
    @pytest.fixture()
    def universe(self):
        return mda.Universe(self.topology, self.filename, mmap=False)

    def test_mmap_kwarg(self, universe):
        assert universe.trajectory._mmap == False


class TestNCDFReader(_NCDFReaderTest, RefVGV):
    pass


class TestNCDFReader_mmap_None(_NCDFReaderTest_mmap_None, RefVGV):
    pass


class TestNCDFReader_mmap_True(_NCDFReaderTest_mmap_True, RefVGV):
    pass


class TestNCDFReader_mmap_False(_NCDFReaderTest_mmap_False, RefVGV):
    pass


class TestNCDFReaderTZ2(_NCDFReaderTest, RefTZ2):
    pass


class TestNCDFReader2(object):
    """NCDF Trajectory with positions and forces.

    Contributed by Albert Solernou
    """

    prec = 3

    @pytest.fixture(scope="class")
    def u(self):
        return mda.Universe(PFncdf_Top, PFncdf_Trj)

    def test_positions_1(self, u):
        """Check positions on first frame"""
        u.trajectory[0]
        ref_1 = np.array(
            [
                [-0.11980818, 18.70524979, 11.6477766],
                [-0.44717646, 18.61727142, 12.59919548],
                [-0.60952115, 19.47885513, 11.22137547],
            ],
            dtype=np.float32,
        )
        assert_almost_equal(ref_1, u.atoms.positions[:3], self.prec)

    def test_positions_2(self, u):
        """Check positions on second frame"""
        u.trajectory[1]
        ref_2 = np.array(
            [
                [-0.13042036, 18.6671524, 11.69647026],
                [-0.46643803, 18.60186768, 12.646698],
                [-0.46567637, 19.49173927, 11.21922874],
            ],
            dtype=np.float32,
        )
        assert_almost_equal(ref_2, u.atoms.positions[:3], self.prec)

    def test_forces_1(self, u):
        """Check forces on first frame"""
        u.trajectory[0]
        ref_1 = np.array(
            [
                [49.23017883, -97.05565643, -86.09863281],
                [2.97547197, 29.84169388, 11.12069607],
                [-15.93093777, 14.43616867, 30.25889015],
            ],
            dtype=np.float32,
        )
        assert_almost_equal(ref_1, u.atoms.forces[:3], self.prec)

    def test_forces_2(self, u):
        """Check forces on second frame"""
        u.trajectory[1]
        ref_2 = np.array(
            [
                [116.39096832, -145.44448853, -151.3155365],
                [-18.90058327, 27.20145798, 1.95245135],
                [-31.08556366, 14.95863628, 41.10367966],
            ],
            dtype=np.float32,
        )
        assert_almost_equal(ref_2, u.atoms.forces[:3], self.prec)

    def test_time_1(self, u):
        """Check time on first frame"""
        ref = 35.02
        assert_almost_equal(ref, u.trajectory[0].time, self.prec)

    def test_time_2(self, u):
        """Check time on second frame"""
        ref = 35.04
        assert_almost_equal(ref, u.trajectory[1].time, self.prec)

    def test_dt(self, u):
        ref = 0.02
        assert_almost_equal(ref, u.trajectory.dt, self.prec)
        assert_almost_equal(ref, u.trajectory.ts.dt, self.prec)

    def test_box(self, u):
        for ts in u.trajectory:
            assert ts.dimensions is None


class TestNCDFReader3(object):
    """NCDF trajectory with box, positions, forces and velocities

    Added to address Issue #2323
    """

    prec = 3

    # Expected coordinates as stored in Angstrom units
    coord_refs = np.array(
        [
            [
                [15.249873, 12.578178, 15.191731],
                [14.925511, 13.58888, 14.944009],
                [15.285703, 14.3409605, 15.645962],
            ],
            [
                [14.799454, 15.214347, 14.714555],
                [15.001984, 15.870884, 13.868363],
                [16.03358, 16.183628, 14.02995],
            ],
        ],
        dtype=np.float32,
    )

    # Expected forces as stored in kcal/(mol*Angstrom)
    frc_refs = np.array(
        [
            [
                [8.583388, 1.8023694, -15.0033455],
                [-21.594835, 39.09166, 6.567963],
                [4.363016, -12.135163, 4.4775457],
            ],
            [
                [-10.106646, -7.870829, -10.385734],
                [7.23599, -12.366022, -9.106191],
                [-4.637955, 11.597565, -6.463743],
            ],
        ],
        dtype=np.float32,
    )

    # Expected velocities as stored in Angstrom per AKMA time unit
    # These are usually associated with a scale_factor of 20.455
    vel_refs = np.array(
        [
            [
                [-0.5301689, -0.16311595, -0.31390688],
                [0.00188578, 0.02513031, -0.2687525],
                [0.84072256, 0.09402391, -0.7457009],
            ],
            [
                [-1.7773226, 1.2307, 0.50276583],
                [-0.13532305, 0.1355039, -0.05567304],
                [-0.6182481, 1.6396415, 0.46686798],
            ],
        ],
        dtype=np.float32,
    )

    # Expected box values as stored in cell_lengths ([:3]) of Angstrom
    # and cell_angles ([:3]) of degree
    box_refs = np.array(
        [
            [28.81876287, 28.27875261, 27.72616397, 90.0, 90.0, 90.0],
            [27.06266081, 26.55555665, 26.03664058, 90.0, 90.0, 90.0],
        ],
        dtype=np.float32,
    )

    @pytest.fixture(scope="class")
    def universe(self):
        return mda.Universe(PRM_NCBOX, TRJ_NCBOX)

    @pytest.mark.parametrize("index,expected", ((0, 0), (8, 1)))
    def test_positions(self, universe, index, expected):
        universe.trajectory[index]
        assert_almost_equal(
            self.coord_refs[expected], universe.atoms.positions[:3], self.prec
        )

    @pytest.mark.parametrize("index,expected", ((0, 0), (8, 1)))
    def test_forces(self, universe, index, expected):
        """Here we multiply the forces by 4.184 to convert from
        kcal to kj in order to verify that MDA has correctly read
        and converted the units from those stored in the NetCDF file.
        """
        universe.trajectory[index]
        assert_almost_equal(
            self.frc_refs[expected] * 4.184,
            universe.atoms.forces[:3],
            self.prec,
        )

    @pytest.mark.parametrize("index,expected", ((0, 0), (8, 1)))
    def test_velocities(self, universe, index, expected):
        """Here we multiply the velocities by 20.455 to match the value of
        `scale_factor` which has been declared in the NetCDF file, which
        should change the values from Angstrom/AKMA time unit to Angstrom/ps.
        """
        universe.trajectory[index]
        assert_almost_equal(
            self.vel_refs[expected] * 20.455,
            universe.atoms.velocities[:3],
            self.prec,
        )

    @pytest.mark.parametrize("index,expected", ((0, 1.0), (8, 9.0)))
    def test_time(self, universe, index, expected):
        assert_almost_equal(
            expected, universe.trajectory[index].time, self.prec
        )

    def test_nframes(self, universe):
        assert_equal(10, universe.trajectory.n_frames)

    def test_dt(self, universe):
        ref = 1.0
        assert_almost_equal(ref, universe.trajectory.dt, self.prec)
        assert_almost_equal(ref, universe.trajectory.ts.dt, self.prec)

    @pytest.mark.parametrize("index,expected", ((0, 0), (8, 1)))
    def test_box(self, universe, index, expected):
        universe.trajectory[index]
        assert_almost_equal(self.box_refs[expected], universe.dimensions)


class TestNCDFReader4(object):
    """NCDF Trajectory exported by cpptaj, without `time` variable."""

    prec = 3

    @pytest.fixture(scope="class")
    def u(self):
        return mda.Universe(CPPTRAJ_TRAJ_TOP, [CPPTRAJ_TRAJ, CPPTRAJ_TRAJ])

    def test_chain_times(self, u):
        """Check times entries for a chain of trajectories without
        a defined time variable"""
        ref_times = [0.0, 1.0, 2.0, 3.0, 4.0, 5.0]
        time_list = [ts.time for ts in u.trajectory]
        assert ref_times == time_list

    def test_dt(self, u):
        ref = 1.0
        assert u.trajectory.dt == pytest.approx(ref)
        assert u.trajectory.ts.dt == pytest.approx(ref)

    def test_warn_user_no_time_information(self, u):
        wmsg = (
            "NCDF trajectory does not contain `time` information;"
            " `time` will be set as an increasing index"
        )
        with pytest.warns(UserWarning, match=wmsg):
            u2 = mda.Universe(CPPTRAJ_TRAJ_TOP, CPPTRAJ_TRAJ)


class _NCDFGenerator(object):
    """A class for generating abitrary ncdf files and exhaustively test
    edge cases which might not be found in the wild"""

    def create_ncdf(self, params):
        """A basic modular ncdf writer based on :class:`NCDFWriter`"""
        # Create under context manager
        with netcdf_file(
            params["filename"], mode="w", version=params["version_byte"]
        ) as ncdf:
            # Top level attributes
            if params["Conventions"]:
                setattr(ncdf, "Conventions", params["Conventions"])
            if params["ConventionVersion"]:
                setattr(ncdf, "ConventionVersion", params["ConventionVersion"])
            if params["program"]:
                setattr(ncdf, "program", params["program"])
            if params["programVersion"]:
                setattr(ncdf, "programVersion", params["programVersion"])

            # Dimensions
            if params["frame"]:
                ncdf.createDimension("frame", None)
            if params["n_atoms"]:
                ncdf.createDimension("atom", params["n_atoms"])
            if params["spatial"]:
                ncdf.createDimension("spatial", params["spatial"])
            if params["time"]:
                ncdf.createDimension("time", 1)
            ncdf.createDimension("label", 5)
            ncdf.createDimension("cell_spatial", 3)
            ncdf.createDimension("cell_angular", 3)

            # Variables
            if params["time"]:
                time = ncdf.createVariable("time", "d", ("time",))
                setattr(time, "units", params["time"])
                time[:] = 1.0
            cell_spatial = ncdf.createVariable(
                "cell_spatial", "c", ("cell_spatial",)
            )
            cell_spatial[:] = np.asarray(list("abc"))
            cell_angular = ncdf.createVariable(
                "cell_angular", "c", ("cell_angular", "label")
            )
            cell_angular[:] = np.asarray(
                [list("alpha"), list("beta "), list("gamma")]
            )

            # Spatial or atom dependent variables
            if (params["spatial"]) and (params["n_atoms"]):
                spatial = ncdf.createVariable("spatial", "c", ("spatial",))
                spatial[:] = np.asarray(list("xyz")[: params["spatial"]])
                if params["frame"]:
                    if params["coordinates"]:
                        coords = ncdf.createVariable(
                            "coordinates", "f4", ("frame", "atom", "spatial")
                        )
                    velocs = ncdf.createVariable(
                        "velocities", "f4", ("frame", "atom", "spatial")
                    )
                    forces = ncdf.createVariable(
                        "forces", "f4", ("frame", "atom", "spatial")
                    )
                    cell_lengths = ncdf.createVariable(
                        "cell_lengths", "f8", ("frame", "cell_spatial")
                    )
                    cell_angles = ncdf.createVariable(
                        "cell_angles", "f8", ("frame", "cell_angular")
                    )
                else:
                    if params["coordinates"]:
                        coords = ncdf.createVariable(
                            "coordinates", "f8", ("atom", "spatial")
                        )
                    cell_lengths = ncdf.createVariable(
                        "cell_lengths", "f8", ("cell_spatial",)
                    )
                    cell_angles = ncdf.createVariable(
                        "cell_angles", "f8", ("cell_angular",)
                    )
                    velocs = ncdf.createVariable(
                        "velocities", "f8", ("atom", "spatial")
                    )
                    forces = ncdf.createVariable(
                        "forces", "f8", ("atom", "spatial")
                    )

                # Set units
                if params["coordinates"]:
                    setattr(coords, "units", params["coordinates"])
                setattr(velocs, "units", params["velocities"])
                setattr(forces, "units", params["forces"])
                setattr(cell_lengths, "units", params["cell_lengths"])
                setattr(cell_angles, "units", params["cell_angles"])

                # Assign value
                if params["frame"]:
                    for index in range(params["frame"]):
                        if params["coordinates"]:
                            coords[index, :] = np.asarray(
                                range(params["spatial"]), dtype=np.float32
                            )
                        cell_lengths[index, :] = np.array(
                            [20.0, 20.0, 20.0], dtype=np.float32
                        )
                        cell_angles[index, :] = np.array(
                            [90.0, 90.0, 90.0], dtype=np.float32
                        )
                        velocs[index, :] = np.asarray(
                            range(params["spatial"]), dtype=np.float32
                        )
                        forces[index, :] = np.asarray(
                            range(params["spatial"]), dtype=np.float32
                        )
                else:
                    if params["coordinates"]:
                        coords[:] = np.asarray(
                            range(params["spatial"]), dtype=np.float32
                        )
                    cell_lengths[:] = np.array(
                        [20.0, 20.0, 20.0], dtype=np.float32
                    )
                    cell_angles[:] = np.array(
                        [90.0, 90.0, 90.0], dtype=np.float32
                    )
                    velocs[:] = np.asarray(
                        range(params["spatial"]), dtype=np.float32
                    )
                    forces[:] = np.asarray(
                        range(params["spatial"]), dtype=np.float32
                    )

            # self.scale_factor overrides which variable gets a scale_factor
            if params["scale_factor"]:
                setattr(
                    ncdf.variables[params["scale_factor"]],
                    "scale_factor",
                    params["scale_factor_value"],
                )

    def gen_params(self, keypair=None, restart=False):
        """Generate writer parameters, keypair can be used to overwrite
        given dictonary entries (expects dictionary)
        """

        params = {
            "filename": "test.nc",
            "version_byte": 2,
            "Conventions": "AMBER",
            "ConventionVersion": "1.0",
            "program": "mda test_writer",
            "programVersion": "V42",
            "n_atoms": 1,
            "spatial": 3,
            "coordinates": "angstrom",
            "velocities": "angstrom/picosecond",
            "forces": "kilocalorie/mole/angstrom",
            "cell_lengths": "angstrom",
            "cell_angles": "degree",
            "time": "picosecond",
            "scale_factor": None,
            "scale_factor_value": 2.0,
            "frame": 2,
        }

        if restart:
            params["filename"] = "test.ncrst"
            params["frame"] = None

        if keypair:
            for entry in keypair:
                params[entry] = keypair[entry]

        return params


class TestScaleFactorImplementation(_NCDFGenerator):

    prec = 5

    def test_scale_factor_coordinates(self, tmpdir):
        mutation = {"scale_factor": "coordinates"}
        params = self.gen_params(keypair=mutation, restart=False)
        expected = np.asarray(range(3), dtype=np.float32) * 2.0
        with tmpdir.as_cwd():
            self.create_ncdf(params)
            u = mda.Universe(params["filename"], to_guess=())
            for ts in u.trajectory:
                assert_almost_equal(ts.positions[0], expected, self.prec)

    def test_scale_factor_velocities(self, tmpdir):
        mutation = {"scale_factor": "velocities", "scale_factor_value": 3.0}
        params = self.gen_params(keypair=mutation, restart=False)
        expected = np.asarray(range(3), dtype=np.float32) * 3.0
        with tmpdir.as_cwd():
            self.create_ncdf(params)
            u = mda.Universe(params["filename"], to_guess=())
            for ts in u.trajectory:
                assert_almost_equal(ts.velocities[0], expected, self.prec)

    def test_scale_factor_forces(self, tmpdir):
        mutation = {"scale_factor": "forces", "scale_factor_value": 10.0}
        params = self.gen_params(keypair=mutation, restart=False)
        expected = np.asarray(range(3), dtype=np.float32) * 10.0 * 4.184
        with tmpdir.as_cwd():
            self.create_ncdf(params)
            u = mda.Universe(params["filename"], to_guess=())
            for ts in u.trajectory:
                assert_almost_equal(ts.forces[0], expected, self.prec)

    @pytest.mark.parametrize(
        "mutation,expected",
        (
            (
                {"scale_factor": "cell_lengths", "scale_factor_value": 0.75},
                np.array([15.0, 15.0, 15.0, 90.0, 90.0, 90.0]),
            ),
            (
                {"scale_factor": "cell_angles", "scale_factor_value": 0.5},
                np.array([20.0, 20.0, 20.0, 45.0, 45.0, 45.0]),
            ),
        ),
    )
    def test_scale_factor_box(self, tmpdir, mutation, expected):
        params = self.gen_params(keypair=mutation, restart=False)
        with tmpdir.as_cwd():
            self.create_ncdf(params)
            u = mda.Universe(params["filename"], to_guess=())
            for ts in u.trajectory:
                assert_almost_equal(ts.dimensions, expected, self.prec)

    def test_scale_factor_not_float(self, tmpdir):
        mutation = {
            "scale_factor": "coordinates",
            "scale_factor_value": "parsnips",
        }
        params = self.gen_params(keypair=mutation, restart=False)
        with tmpdir.as_cwd():
            self.create_ncdf(params)
            errmsg = "b'parsnips' is not a float"
            with pytest.raises(TypeError, match=errmsg):
                u = mda.Universe(params["filename"])


class TestNCDFReaderExceptionsWarnings(_NCDFGenerator):

    @pytest.mark.parametrize(
        "mutation",
        [{"Conventions": "Foo"}, {"version_byte": 1}, {"spatial": 2}],
    )
    def test_type_errors(self, tmpdir, mutation):
        params = self.gen_params(keypair=mutation, restart=False)
        with tmpdir.as_cwd():
            self.create_ncdf(params)
            with pytest.raises(TypeError):
                NCDFReader(params["filename"])

    @pytest.mark.parametrize(
        "mutation",
        [
            {"Conventions": None},
            {"ConventionVersion": None},
            {"spatial": None},
            {"n_atoms": None},
            {"frame": None},
        ],
    )
    def test_value_errors(self, tmpdir, mutation):
        params = self.gen_params(keypair=mutation, restart=False)
        with tmpdir.as_cwd():
            self.create_ncdf(params)
            with pytest.raises(ValueError):
                NCDFReader(params["filename"])

    @pytest.mark.parametrize(
        "mutation",
        [
            {"scale_factor": "cell_spatial"},
            {"time": "femtosecond"},
            {"coordinates": "nanometer"},
            {"velocities": "angstrom/akma"},
            {"forces": "kilojoule/mole/angstrom"},
            {"cell_lengths": "nanometer"},
            {"cell_angles": "radians"},
        ],
    )
    def test_notimplemented_errors(self, tmpdir, mutation):
        params = self.gen_params(keypair=mutation, restart=False)
        with tmpdir.as_cwd():
            self.create_ncdf(params)
            with pytest.raises(NotImplementedError):
                NCDFReader(params["filename"])

    @pytest.mark.parametrize(
        "evaluate,expected", (("yard", "foot"), ("second", "minute"))
    )
    def test_verify_units_errors(self, evaluate, expected):
        """Directly tests expected failures of _verify_units"""
        with pytest.raises(NotImplementedError):
            NCDFReader._verify_units(evaluate.encode("utf-8"), expected)

    def test_ioerror(self, tmpdir):
        params = self.gen_params(restart=False)
        with tmpdir.as_cwd():
            self.create_ncdf(params)
            with pytest.raises(IOError):
                u = mda.Universe(params["filename"], to_guess=())
                u.trajectory.close()
                u.trajectory[-1]

    def test_conventionversion_warn(self, tmpdir):
        mutation = {"ConventionVersion": "2.0"}
        params = self.gen_params(keypair=mutation, restart=False)
        with tmpdir.as_cwd():
            self.create_ncdf(params)
            with pytest.warns(UserWarning) as record:
                NCDFReader(params["filename"])

            assert len(record) == 1
            wmsg = (
                "NCDF trajectory format is 2.0 but the reader "
                "implements format 1.0"
            )
            assert str(record[0].message.args[0]) == wmsg

    @pytest.mark.parametrize(
        "mutation", [{"program": None}, {"programVersion": None}]
    )
    def test_program_warn(self, tmpdir, mutation):
        params = self.gen_params(keypair=mutation, restart=False)
        with tmpdir.as_cwd():
            self.create_ncdf(params)
            with pytest.warns(UserWarning) as record:
                NCDFReader(params["filename"])

            assert len(record) == 1
            wmsg = (
                "NCDF trajectory test.nc may not fully adhere to AMBER "
                "standards as either the `program` or `programVersion` "
                "attributes are missing"
            )
            assert str(record[0].message.args[0]) == wmsg

    def test_no_dt_warning(self, tmpdir):
        """Issue 3166 - not being able to call dt should throw a warning.
        Also at the same time checks that single frame chain reading works"""
        u = mda.Universe(PFncdf_Top, PFncdf_Trj)
        with tmpdir.as_cwd():
            with NCDFWriter("single_frame.nc", u.atoms.n_atoms) as W:
                W.write(u)

            # Using the ChainReader implicitly calls dt() and thus _get_dt()
            wmsg = "Reader has no dt information, set to 1.0 ps"
            with pytest.warns(UserWarning, match=wmsg):
                u2 = mda.Universe(PFncdf_Top, [PFncdf_Trj, "single_frame.nc"])


class _NCDFWriterTest(object):
    prec = 5

    @pytest.fixture()
    def universe(self):
        return mda.Universe(self.topology, self.filename)

    @pytest.fixture(params=["nc", "ncdf"])
    def outfile_extensions(self, tmpdir, request):
        # Issue 3030, test all extensions of NCDFWriter
        ext = request.param
        return str(tmpdir) + f"ncdf-writer-1.{ext}"

    @pytest.fixture()
    def outfile(self, tmpdir):
        return str(tmpdir) + "ncdf-writer-1.ncdf"

    @pytest.fixture()
    def outtop(self, tmpdir):
        return str(tmpdir) + "ncdf-writer-top.pdb"

    def _test_write_trajectory(self, universe, outfile):
        # explicit import so that we can artifically remove netCDF4
        # before calling
        from MDAnalysis.coordinates import TRJ

        t = universe.trajectory
        with TRJ.NCDFWriter(outfile, t.n_atoms, dt=t.dt) as W:
            self._copy_traj(W, universe)
        self._check_new_traj(universe, outfile)
        # for issue #518 -- preserve float32 data in ncdf output
        # NOTE: This originally failed with the dtype('>f4') instead
        #       of dtype('<f4') == dtype('f') == np.float32, i.e. then
        #       endianness is different. The current hack-ish solution
        #       ignores endianness by comparing the name of the types,
        #       which should be "float32".
        #       See http://docs.scipy.org/doc/numpy-1.10.0/reference/arrays.dtypes.html
        #       and https://github.com/MDAnalysis/mdanalysis/pull/503
        dataset = netcdf_file(outfile, "r")
        coords = dataset.variables["coordinates"]
        time = dataset.variables["time"]
        assert_equal(
            coords[:].dtype.name,
            np.dtype(np.float32).name,
            err_msg="ncdf coord output not float32 "
            "but {}".format(coords[:].dtype),
        )
        assert_equal(
            time[:].dtype.name,
            np.dtype(np.float32).name,
            err_msg="ncdf time output not float32 "
            "but {}".format(time[:].dtype),
        )

    def test_write_trajectory_netCDF4(self, universe, outfile):
        pytest.importorskip("netCDF4")
        return self._test_write_trajectory(universe, outfile)

    def test_write_trajectory_netcdf(self, universe, outfile):
        import MDAnalysis.coordinates.TRJ

        loaded_netCDF4 = sys.modules["MDAnalysis.coordinates.TRJ"].netCDF4
        try:
            # cannot use @block_import('netCDF4') because TRJ was already imported
            # during setup() and already sits in the global module list so we just
            # set it to None because that is what TRJ does if it cannot find netCDF4
            sys.modules["MDAnalysis.coordinates.TRJ"].netCDF4 = None
            assert (
                MDAnalysis.coordinates.TRJ.netCDF4 is None
            )  # should happen if netCDF4 not found
            return self._test_write_trajectory(universe, outfile)
        finally:
            sys.modules["MDAnalysis.coordinates.TRJ"].netCDF4 = loaded_netCDF4

    def test_OtherWriter(self, universe, outfile_extensions):
        t = universe.trajectory
        with t.OtherWriter(outfile_extensions) as W:
            self._copy_traj(W, universe)
        self._check_new_traj(universe, outfile_extensions)

    def _copy_traj(self, writer, universe):
        for ts in universe.trajectory:
            writer.write(universe)

    def _check_new_traj(self, universe, outfile):
        uw = mda.Universe(self.topology, outfile)

        # check that the trajectories are identical for each time step
        for orig_ts, written_ts in zip(universe.trajectory, uw.trajectory):
            assert_almost_equal(
                written_ts._pos,
                orig_ts._pos,
                self.prec,
                err_msg="coordinate mismatch between "
                "original and written trajectory at "
                "frame %d (orig) vs %d (written)"
                % (orig_ts.frame, written_ts.frame),
            )
            # not a good test because in the example trajectory all times are 0
            assert_almost_equal(
                orig_ts.time,
                written_ts.time,
                self.prec,
                err_msg="Time for step {0} are not the "
                "same.".format(orig_ts.frame),
            )
            assert_almost_equal(
                written_ts.dimensions,
                orig_ts.dimensions,
                self.prec,
                err_msg="unitcells are not identical",
            )
        # check that the NCDF data structures are the same
        nc_orig = universe.trajectory.trjfile
        nc_copy = uw.trajectory.trjfile

        # note that here 'dimensions' is a specific netcdf data structure and
        # not the unit cell dimensions in MDAnalysis
        for k, dim in nc_orig.dimensions.items():
            try:
                dim_new = nc_copy.dimensions[k]
            except KeyError:
                raise AssertionError(
                    "NCDFWriter did not write " "dimension '{0}'".format(k)
                )
            else:
                assert_equal(
                    dim,
                    dim_new,
                    err_msg="Dimension '{0}' size mismatch".format(k),
                )

        for k, v in nc_orig.variables.items():
            try:
                v_new = nc_copy.variables[k]
            except KeyError:
                raise AssertionError(
                    "NCDFWriter did not write " "variable '{0}'".format(k)
                )
            else:
                try:
                    assert_almost_equal(
                        v[:],
                        v_new[:],
                        self.prec,
                        err_msg="Variable '{0}' not "
                        "written correctly".format(k),
                    )
                except TypeError:
                    assert_equal(
                        v[:],
                        v_new[:],
                        err_msg="Variable {0} not written "
                        "correctly".format(k),
                    )

    def test_TRR2NCDF(self, outfile):
        trr = mda.Universe(GRO, TRR)
        with mda.Writer(
            outfile, trr.trajectory.n_atoms, velocities=True, format="ncdf"
        ) as W:
            for ts in trr.trajectory:
                W.write(trr)

        uw = mda.Universe(GRO, outfile)

        for orig_ts, written_ts in zip(trr.trajectory, uw.trajectory):
            assert_almost_equal(
                written_ts._pos,
                orig_ts._pos,
                self.prec,
                err_msg="coordinate mismatch between "
                "original and written trajectory at "
                "frame {0} (orig) vs {1} (written)".format(
                    orig_ts.frame, written_ts.frame
                ),
            )
            assert_almost_equal(
                written_ts._velocities,
                orig_ts._velocities,
                self.prec,
                err_msg="velocity mismatch between "
                "original and written trajectory at "
                "frame {0} (orig) vs {1} (written)".format(
                    orig_ts.frame, written_ts.frame
                ),
            )
            assert_almost_equal(
                orig_ts.time,
                written_ts.time,
                self.prec,
                err_msg="Time for step {0} are not the "
                "same.".format(orig_ts.frame),
            )
            assert_almost_equal(
                written_ts.dimensions,
                orig_ts.dimensions,
                self.prec,
                err_msg="unitcells are not identical",
            )
        del trr

    def test_write_AtomGroup(self, universe, outfile, outtop):
        """test to write NCDF from AtomGroup (Issue 116)"""
        p = universe.select_atoms("not resname WAT")
        p.write(outtop)
        with mda.Writer(outfile, n_atoms=p.n_atoms, format="ncdf") as W:
            for ts in universe.trajectory:
                W.write(p)

        uw = mda.Universe(outtop, outfile)
        pw = uw.atoms

        for orig_ts, written_ts in zip(universe.trajectory, uw.trajectory):
            assert_almost_equal(
                p.positions,
                pw.positions,
                self.prec,
                err_msg="coordinate mismatch between "
                "original and written trajectory at "
                "frame %d (orig) vs %d (written)"
                % (orig_ts.frame, written_ts.frame),
            )
            assert_almost_equal(
                orig_ts.time,
                written_ts.time,
                self.prec,
                err_msg="Time for step {0} are not the "
                "same.".format(orig_ts.frame),
            )
            assert_almost_equal(
                written_ts.dimensions,
                orig_ts.dimensions,
                self.prec,
                err_msg="unitcells are not identical",
            )


class TestNCDFWriter(_NCDFWriterTest, RefVGV):
    pass


class TestNCDFWriterTZ2(_NCDFWriterTest, RefTZ2):
    pass


class TestNCDFWriterVelsForces(object):
    """Test writing NCDF trajectories with a mixture of options"""

    prec = 3
    top = XYZ_mini
    n_atoms = 3

    @pytest.fixture()
    def u1(self):
        u = make_Universe(
            size=(self.n_atoms, 1, 1),
            trajectory=True,
            velocities=True,
            forces=True,
        )
        # Memory reader so changes should be in-place
        u.atoms.velocities += 100
        u.atoms.forces += 200
        return u

    @pytest.fixture()
    def u2(self):
        u = make_Universe(
            size=(self.n_atoms, 1, 1),
            trajectory=True,
            velocities=True,
            forces=True,
        )
        # Memory reader so changes should be in-place
        u.atoms.positions += 300
        u.atoms.velocities += 400
        u.atoms.forces += 500
        return u

    @pytest.mark.parametrize(
        "pos, vel, force",
        (
            (True, False, False),
            (True, True, False),
            (True, False, True),
            (True, True, True),
        ),
    )
    def test_write_u(self, pos, vel, force, tmpdir, u1, u2):
        """Write the two reference universes, then open them up and check values

        pos vel and force are bools which define whether these properties
        should be in universe
        """
        outfile = str(tmpdir) + "ncdf-write-vels-force.ncdf"
        with NCDFWriter(
            outfile, n_atoms=self.n_atoms, velocities=vel, forces=force
        ) as w:
            w.write(u1)
            w.write(u2)

        # test that the two reference states differ
        for ts1, ts2 in zip(u1.trajectory, u2.trajectory):
            assert_almost_equal(ts1._pos + 300, ts2._pos)
            assert_almost_equal(ts1._velocities + 300, ts2._velocities)
            assert_almost_equal(ts1._forces + 300, ts2._forces)

        u = mda.Universe(self.top, outfile)
        # check the trajectory contents match reference universes
        for ts, ref_ts in zip(
            u.trajectory, [u1.trajectory.ts, u2.trajectory.ts]
        ):
            if pos:
                assert_almost_equal(ts._pos, ref_ts._pos, self.prec)
            else:
                with pytest.raises(mda.NoDataError):
                    getattr(ts, "positions")
            if vel:
                assert_almost_equal(
                    ts._velocities, ref_ts._velocities, self.prec
                )
            else:
                with pytest.raises(mda.NoDataError):
                    getattr(ts, "velocities")
            if force:
                assert_almost_equal(ts._forces, ref_ts._forces, self.prec)
            else:
                with pytest.raises(mda.NoDataError):
                    getattr(ts, "forces")

        u.trajectory.close()


class TestNCDFWriterScaleFactors:
    """Class to check that the NCDF Writer properly handles the
    application of scale factors"""

    @pytest.fixture()
    def outfile(self, tmpdir):
        return str(tmpdir) + "ncdf-write-scale.ncdf"

    @pytest.fixture()
    def outfile2(self, tmpdir):
        return str(tmpdir) + "ncdf-write-scale2.ncdf"

    @pytest.fixture()
    def universe(self):
        return mda.Universe(PRM_NCBOX, TRJ_NCBOX)

    def get_scale_factors(self, ncdfile):
        """Get a dictionary of scale factors stored in netcdf file"""
        sfactors = {}
        # being overly cautious by setting mmap to False, probably would
        # be faster & ok to set it to True
        with netcdf_file(ncdfile, mmap=False) as f:
            for var in f.variables:
                if hasattr(f.variables[var], "scale_factor"):
                    sfactors[var] = f.variables[var].scale_factor

        return sfactors

    def get_variable(self, ncdfile, variable, frame):
        """Return a variable array from netcdf file"""
        with netcdf_file(ncdfile, mmap=False) as f:
            return f.variables[variable][frame]

    def test_write_read_factors_default(self, outfile, universe):
        with universe.trajectory.Writer(outfile) as W:
            W.write(universe.atoms)

        # check scale_factors
        sfactors = self.get_scale_factors(outfile)
        assert len(sfactors) == 1
        assert sfactors["velocities"] == 20.455

    def test_write_bad_scale_factor(self, outfile, universe):
        errmsg = "scale_factor parsnips is not a float"
        with pytest.raises(TypeError, match=errmsg):
            NCDFWriter(
                outfile,
                n_atoms=len(universe.atoms),
                scale_velocities="parsnips",
            )

    @pytest.mark.parametrize(
        "stime, slengths, sangles, scoords, svels, sfrcs",
        (
            (-2.0, -2.0, -2.0, -2.0, -2.0, -2.0),
            (1.0, 1.0, 1.0, 1.0, 1.0, 1.0),
            (2.0, 4.0, 8.0, 16.0, 32.0, 64.0),
        ),
    )
    def test_write_read_write(
        self,
        outfile,
        outfile2,
        universe,
        stime,
        slengths,
        sangles,
        scoords,
        svels,
        sfrcs,
    ):
        """Write out a file with assorted scale_factors, then
        read it back in, then write it out to make sure that the
        new assorted scale_factors have been retained by Write"""

        with NCDFWriter(
            outfile,
            n_atoms=len(universe.atoms),
            velocities=True,
            forces=True,
            scale_time=stime,
            scale_cell_lengths=slengths,
            scale_cell_angles=sangles,
            scale_coordinates=scoords,
            scale_velocities=svels,
            scale_forces=sfrcs,
        ) as W:
            for ts in universe.trajectory:
                W.write(universe.atoms)

        universe2 = mda.Universe(PRM_NCBOX, outfile)

        # Write back checking that the same scale_factors as used
        with universe2.trajectory.Writer(outfile2) as OWriter:
            OWriter.write(universe2.atoms)

        universe3 = mda.Universe(PRM_NCBOX, outfile2)

        # check stored scale_factors
        sfactors1 = self.get_scale_factors(outfile)
        sfactors2 = self.get_scale_factors(outfile2)

        assert sfactors1 == sfactors2
        assert len(sfactors1) == 6
        assert sfactors1["time"] == stime
        assert sfactors1["cell_lengths"] == slengths
        assert sfactors1["cell_angles"] == sangles
        assert sfactors1["coordinates"] == scoords
        assert sfactors1["velocities"] == svels
        assert sfactors1["forces"] == sfrcs

        # check that the stored values are indeed scaled
        assert_almost_equal(
            universe.trajectory.time / stime,
            self.get_variable(outfile, "time", 0),
            4,
        )
        assert_almost_equal(
            universe.dimensions[:3] / slengths,
            self.get_variable(outfile, "cell_lengths", 0),
            4,
        )
        assert_almost_equal(
            universe.dimensions[3:] / sangles,
            self.get_variable(outfile, "cell_angles", 0),
            4,
        )
        assert_almost_equal(
            universe.atoms.positions / scoords,
            self.get_variable(outfile, "coordinates", 0),
            4,
        )
        assert_almost_equal(
            universe.atoms.velocities / svels,
            self.get_variable(outfile, "velocities", 0),
            4,
        )
        # note: kJ/mol -> kcal/mol = 4.184 conversion
        assert_almost_equal(
            universe.atoms.forces / (sfrcs * 4.184),
            self.get_variable(outfile, "forces", 0),
            4,
        )

        # check that the individual components were saved/read properly
        for ts1, ts3 in zip(universe.trajectory, universe3.trajectory):
            assert_almost_equal(ts1.time, ts3.time)
            assert_almost_equal(ts1.dimensions, ts3.dimensions)
            assert_almost_equal(
                universe.atoms.positions, universe3.atoms.positions, 4
            )
            assert_almost_equal(
                universe.atoms.velocities, universe3.atoms.velocities, 4
            )
            assert_almost_equal(
                universe.atoms.forces, universe3.atoms.forces, 4
            )


class TestScipyScaleFactors(TestNCDFWriterScaleFactors):
    """As above, but netCDF4 is disabled since scaleandmask is different
    between the two libraries"""

    @pytest.fixture(autouse=True)
    def block_netcdf4(self, monkeypatch):
        monkeypatch.setattr(
            sys.modules["MDAnalysis.coordinates.TRJ"], "netCDF4", None
        )

    def test_ncdf4_not_present(self, outfile, universe):
        # whilst we're here, let's also test this warning
        with pytest.warns(UserWarning, match="Could not find netCDF4 module"):
            with universe.trajectory.Writer(outfile) as W:
                W.write(universe.atoms)


class TestNCDFWriterUnits(object):
    """Tests that the writer adheres to AMBER convention units"""

    @pytest.fixture()
    def outfile(self, tmpdir):
        return str(tmpdir) + "ncdf-writer-1.ncdf"

    @pytest.mark.parametrize(
        "var, expected",
        (
            ("coordinates", "angstrom"),
            ("time", "picosecond"),
            ("cell_lengths", "angstrom"),
            ("cell_angles", "degree"),
            ("velocities", "angstrom/picosecond"),
            ("forces", "kilocalorie/mole/angstrom"),
        ),
    )
    def test_writer_units(self, outfile, var, expected):
        trr = mda.Universe(DLP_CONFIG, format="CONFIG")

        with mda.Writer(
            outfile,
            trr.trajectory.n_atoms,
            velocities=True,
            forces=True,
            format="ncdf",
        ) as W:
            for ts in trr.trajectory:
                W.write(trr)

        with netcdf_file(outfile, mode="r") as ncdf:
            unit = ncdf.variables[var].units.decode("utf-8")
            assert_equal(unit, expected)


class TestNCDFWriterErrorsWarnings(object):
    @pytest.fixture()
    def outfile(self, tmpdir):
        return str(tmpdir) + "out.ncdf"

    def test_zero_atoms_VE(self, outfile):
        with pytest.raises(ValueError):
            NCDFWriter(outfile, 0)

    def test_wrong_n_atoms(self, outfile):
        with NCDFWriter(outfile, 100) as w:
            u = make_Universe(trajectory=True)
            with pytest.raises(IOError):
                w.write(u)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
FHI-AIMS Topology Parser --- :mod:`MDAnalysis.topolgy.FHIAIMSParser`
====================================================================

Reads an `FHI-AIMS`_ ``.in`` file and pulls the atom information from it.
Because an FHI-AIMS input file only has atom name information, any
information about residues and segments will not be populated.

.. _`FHI-AIMS`: https://aimsclub.fhi-berlin.mpg.de/


See Also
--------
:mod:`MDAnalysis.coordinates.FHIAIMS`


Classes
-------

.. autoclass:: FHIAIMSParser
   :members:
   :inherited-members:

"""
import numpy as np

from ..lib.util import openany
from .base import TopologyReaderBase
from ..core.topology import Topology
from ..core.topologyattrs import (
    Atomnames,
    Atomids,
    Resids,
    Resnums,
    Segids,
    Elements,
)


class FHIAIMSParser(TopologyReaderBase):
    """Parse a list of atoms from an FHI-AIMS file

    Creates the following attributes:
     - Atomnames


    .. note::

        By default, atomtypes and masses will be guessed on Universe creation.
        This may change in release 3.0.
        See :ref:`Guessers` for more information.

    .. versionchanged:: 2.8.0
        Removed type and mass guessing (attributes guessing takes place now
        through universe.guess_TopologyAttrs() API).
    """

    format = ["IN", "FHIAIMS"]

    def parse(self, **kwargs):
        """Read the file and return the structure.

        Returns
        -------
        MDAnalysis Topology object
        """
        # FHIAIMS geometry files are only single frames
        names = []
        skip_tags = ["#", "lattice_vector", "initial_moment", "velocity"]
        with openany(self.filename) as inf:
            for line in inf:
                line = line.strip()
                if line.startswith("atom"):
                    names.append(line.split()[-1])
                    continue
                if any([line.startswith(tag) for tag in skip_tags]):
                    continue
                # we are now seeing something that's neither atom nor lattice
                raise ValueError(
                    "Non-conforming line: ({0})in FHI-AIMS input file {0}".format(
                        line, self.filename
                    )
                )
            names = np.asarray(names)
            natoms = len(names)

        attrs = [
            Atomnames(names),
            Atomids(np.arange(natoms) + 1),
            Resids(np.array([1])),
            Resnums(np.array([1])),
            Segids(np.array(["SYSTEM"], dtype=object)),
            Elements(names),
        ]

        top = Topology(natoms, 1, 1, attrs=attrs)

        return top



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
"""
Reading trajectories with *chemfiles* --- :mod:`MDAnalysis.coordinates.chemfiles`
=================================================================================

MDAnalysis interoperates with the `chemfiles`_ library. The *chemfiles* C++ library
supports an expanding set of file formats, some of which are not natively supported by
MDAnalysis. Using the *CHEMFILES* reader you can use  `chemfiles`_ for the low-level
file reading. Check the list of `chemfile-supported file formats`_.

.. _chemfiles: https://chemfiles.org
.. _chemfile-supported file formats: https://chemfiles.org/chemfiles/latest/formats.html#list-of-supported-formats

Using the CHEMFILES reader
--------------------------

When reading, set the ``format="CHEMFILES"`` keyword argument and I/O is delegated to
`chemfiles`_. For example::

   >>> import MDAnalysis as mda
   >>> from MDAnalysis.tests import datafiles as data
   >>> u = mda.Universe(data.TPR, data.TRR, format="CHEMFILES")
   >>> print(u.trajectory)
   <ChemfilesReader ~/anaconda3/envs/mda3/lib/python3.8/site-packages/MDAnalysisTests/data/adk_oplsaa.trr with 10 frames of 47681 atoms>

You can then use the :class:`~MDAnalysis.core.universe.Universe` as usual while chemfiles
is handling the I/O transparently in the background.

`chemfiles`_ can also *write* a number of formats for which there are no Writers in
MDAnalysis. For example, to write a mol2 file::

   >>> u = mda.Universe(data.mol2_ligand)
   >>> with mda.Writer("ligand.mol2", format="CHEMFILES") as W:
   ...     W.write(u.atoms)




Classes
-------

Classes to read and write files using the `chemfiles`_ library. This library
provides C++ implementation of multiple formats readers and writers.

.. autoclass:: ChemfilesReader

.. autoclass:: ChemfilesWriter

.. autoclass:: ChemfilesPicklable

Helper functions
----------------

.. autodata:: MIN_CHEMFILES_VERSION
.. autodata:: MAX_CHEMFILES_VERSION
.. autofunction:: check_chemfiles_version

"""
import numpy as np
import warnings
from packaging.version import Version

from . import base
from ..lib.util import store_init_arguments

try:
    import chemfiles
except ImportError:
    HAS_CHEMFILES = False

    # Allow building documentation even if chemfiles is not installed
    import types

    class MockTrajectory:
        pass

    chemfiles = types.ModuleType("chemfiles")
    chemfiles.Trajectory = MockTrajectory
else:
    HAS_CHEMFILES = True


#: Lowest version of chemfiles that is supported by MDAnalysis.
MIN_CHEMFILES_VERSION = Version("0.10")
#: Lowest version of chemfiles that is *not supported* by MDAnalysis.
MAX_CHEMFILES_VERSION = Version("0.11")


def check_chemfiles_version():
    """Check if an appropriate *chemfiles* is available

    Returns ``True`` if a usable chemfiles version is available,
    with :data:`MIN_CHEMFILES_VERSION` <= version <
    :data:`MAX_CHEMFILES_VERSION`

    .. versionadded:: 1.0.0
    """
    if not HAS_CHEMFILES:
        warnings.warn(
            "chemfiles package not found, "
            "try installing it with 'pip install chemfiles'"
        )
        return False
    version = Version(chemfiles.__version__)
    return version >= MIN_CHEMFILES_VERSION and version < MAX_CHEMFILES_VERSION


class ChemfilesReader(base.ReaderBase):
    """Coordinate reader using chemfiles.

    The file format to used is guessed based on the file extension. If no
    matching format is found, a ``ChemfilesError`` is raised. It is also
    possible to manually specify the format to use for a given file.

    .. versionadded:: 1.0.0
    """

    format = "chemfiles"
    units = {"time": "fs", "length": "Angstrom"}

    @store_init_arguments
    def __init__(self, filename, chemfiles_format="", **kwargs):
        """
        Parameters
        ----------
        filename : chemfiles.Trajectory or str
            the chemfiles object to read or filename to read
        chemfiles_format : str (optional)
            if *filename* was a string, use the given format name instead of
            guessing from the extension. The `list of supported formats
            <formats>`_ and the associated names is available in the chemfiles
            documentation.
        **kwargs : dict
            General reader arguments.

        """
        if not check_chemfiles_version():
            raise RuntimeError(
                "Please install Chemfiles > {}"
                "".format(MIN_CHEMFILES_VERSION)
            )
        super(ChemfilesReader, self).__init__(filename, **kwargs)
        self._format = chemfiles_format
        self._cached_n_atoms = None
        self._open()
        self.ts = self._Timestep(self.n_atoms, **self._ts_kwargs)
        self.next()

    @staticmethod
    def _format_hint(thing):
        """Can this Reader read *thing*"""
        # nb, filename strings can still get passed through if
        # format='CHEMFILES' is used
        return HAS_CHEMFILES and isinstance(thing, chemfiles.Trajectory)

    def _open(self):
        self._closed = False
        self._step = 0
        self._frame = -1
        if isinstance(self.filename, chemfiles.Trajectory):
            self._file = self.filename
        else:
            self._file = ChemfilesPicklable(self.filename, "r", self._format)

    def close(self):
        """close reader"""
        if not self._closed:
            self._file.close()
            self._closed = True

    @property
    def n_frames(self):
        """number of frames in trajectory"""
        return self._file.nsteps

    @property
    def n_atoms(self):
        """number of atoms in the first frame of the trajectory"""
        if self._cached_n_atoms is None:
            self._cached_n_atoms = len(self._file.read_step(0).atoms)
        return self._cached_n_atoms

    def _reopen(self):
        """reopen trajectory"""
        self.close()
        self._open()

    def _read_frame(self, i):
        """read frame i"""
        self._step = i
        return self._read_next_timestep()

    def _read_next_timestep(self, ts=None):
        """copy next frame into timestep"""
        if self._step >= self.n_frames:
            raise IOError("trying to go over trajectory limit")
        if ts is None:
            ts = self.ts
        self.ts = ts
        frame = self._file.read_step(self._step)
        self._frame_to_ts(frame, ts)
        ts.frame = frame.step
        self._step += 1
        return ts

    def _frame_to_ts(self, frame, ts):
        """convert a chemfiles frame to a :class:`TimeStep`"""
        if len(frame.atoms) != self.n_atoms:
            raise IOError(
                "The number of atom changed in the trajectory. "
                "This is not supported by MDAnalysis."
            )

        ts.dimensions = np.r_[frame.cell.lengths, frame.cell.angles]
        ts.positions[:] = frame.positions[:]
        if frame.has_velocities():
            ts.has_velocities = True
            ts.velocities[:] = frame.velocities[:]

    def Writer(self, filename, n_atoms=None, **kwargs):
        """Return writer for trajectory format"""
        if n_atoms is None:
            n_atoms = self.n_atoms
        return ChemfilesWriter(filename, n_atoms, **kwargs)


class ChemfilesWriter(base.WriterBase):
    """
    Coordinate writer using chemfiles.

    The file format to used is guessed based on the file extension. If no
    matching format is found, a ``ChemfilesError`` is raised. It is also
    possible to manually specify the format to use for a given file.

    Chemfiles support writting to files with varying number of atoms if the
    underlying format support it. This is the case of most of text-based
    formats.

    .. versionadded:: 1.0.0
    """

    format = "chemfiles"
    multiframe = True

    # chemfiles mostly[1] uses these units for the in-memory representation,
    # and convert into the file format units when writing.
    #
    # [1] mostly since some format don't have a specified unit
    # (XYZ for example), so then chemfiles just assume they are in A and fs.
    units = {"time": "fs", "length": "Angstrom"}

    def __init__(
        self,
        filename,
        n_atoms=0,
        mode="w",
        chemfiles_format="",
        topology=None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        filename: str
            filename of trajectory file.
        n_atoms: int
            number of atoms in the trajectory to write. This value is not
            used and can vary during trajectory, if the underlying format
            support it
        mode : str (optional)
            file opening mode: use 'a' to append to an existing file or 'w' to
            create a new file
        chemfiles_format : str (optional)
            use the given format name instead of guessing from the extension.
            The `list of supported formats <formats>`_ and the associated names
            is available in chemfiles documentation.
        topology : Universe or AtomGroup (optional)
            use the topology from this :class:`~MDAnalysis.core.groups.AtomGroup`
            or :class:`~MDAnalysis.core.universe.Universe` to write all the
            timesteps to the file
        **kwargs : dict
            General writer arguments.

        """
        if not check_chemfiles_version():
            raise RuntimeError(
                "Please install Chemfiles > {}"
                "".format(MIN_CHEMFILES_VERSION)
            )
        self.filename = filename
        self.n_atoms = n_atoms
        if mode != "a" and mode != "w":
            raise IOError("Expected 'a' or 'w' as mode in ChemfilesWriter")
        self._file = chemfiles.Trajectory(
            self.filename, mode, chemfiles_format
        )
        self._closed = False
        if topology is not None:
            if isinstance(topology, str):
                self._file.set_topology(topology)
            else:
                topology = self._topology_to_chemfiles(topology, n_atoms)
                self._file.set_topology(topology)

    def close(self):
        """Close the trajectory file and finalize the writing"""
        if hasattr(self, "_closed") and not self._closed:
            self._file.close()
            self._closed = True

    def _write_next_frame(self, obj):
        """Write information associated with ``obj`` at current frame into
        trajectory.

        Topology for the output is taken from the `obj` or default to the value
        of the `topology` keyword supplied to the :class:`ChemfilesWriter`
        constructor.

        If `obj` contains velocities, and the underlying format supports it, the
        velocities are written to the file. Writing forces is unsupported at the
        moment.

        Parameters
        ----------
        obj : AtomGroup or Universe
            The :class:`~MDAnalysis.core.groups.AtomGroup` or
            :class:`~MDAnalysis.core.universe.Universe` to write.


        .. versionchanged:: 2.0.0
           Deprecated support for Timestep argument has now been removed.
           Use AtomGroup or Universe as an input instead.
        """
        try:
            atoms = obj.atoms
        except AttributeError:
            errmsg = "Input obj is neither an AtomGroup or Universe"
            raise TypeError(errmsg) from None
        else:
            if hasattr(obj, "universe"):
                # For AtomGroup and children (Residue, ResidueGroup, Segment)
                ts_full = obj.universe.trajectory.ts
                if ts_full.n_atoms == atoms.n_atoms:
                    ts = ts_full
                else:
                    # Only populate a time step with the selected atoms.
                    ts = ts_full.copy_slice(atoms.indices)
            elif hasattr(obj, "trajectory"):
                # For Universe only --- get everything
                ts = obj.trajectory.ts

        frame = self._timestep_to_chemfiles(ts)
        frame.topology = self._topology_to_chemfiles(obj, len(frame.atoms))

        self._file.write(frame)

    def _timestep_to_chemfiles(self, ts):
        """
        Convert a Timestep to a chemfiles Frame
        """
        # TODO: CONVERTERS?
        frame = chemfiles.Frame()
        frame.resize(ts.n_atoms)
        if ts.has_positions:
            frame.positions[:] = ts.positions[:]
        if ts.has_velocities:
            frame.add_velocities()
            frame.velocities[:] = ts.velocities[:]

        # if there is no cell information in this universe, still pass a valid
        # cell to chemfiles
        if ts.dimensions is None:
            lengths, angles = np.zeros(3), np.array([90, 90, 90.0])
        else:
            lengths = ts.dimensions[:3]
            angles = ts.dimensions[3:]

        frame.cell = chemfiles.UnitCell(lengths, angles)

        return frame

    def _topology_to_chemfiles(self, obj, n_atoms):
        """
        Convert an AtomGroup or an Universe to a chemfiles Topology
        """
        topology = chemfiles.Topology()
        if not hasattr(obj, "atoms"):
            # use an empty topology
            topology.resize(n_atoms)
            return topology

        # (1) add all atoms to the topology
        residues = {}
        for atom in obj.atoms:
            name = getattr(atom, "name", "")
            type = getattr(atom, "type", name)
            chemfiles_atom = chemfiles.Atom(name, type)

            if hasattr(atom, "altLoc"):
                chemfiles_atom["altloc"] = str(atom.altLoc)

            if hasattr(atom, "segid"):
                chemfiles_atom["segid"] = str(atom.segid)

            if hasattr(atom, "segindex"):
                chemfiles_atom["segindex"] = int(atom.segindex)

            if hasattr(atom, "resid"):
                resname = getattr(atom, "resname", "")
                if atom.resid not in residues.keys():
                    residues[atom.resid] = chemfiles.Residue(
                        resname, atom.resid
                    )
                residue = residues[atom.resid]

                atom_idx = len(topology.atoms)
                residue.atoms.append(atom_idx)

                if hasattr(atom, "record_type"):
                    # set corresponding chemfiles residue property
                    if atom.record_type == "ATOM":
                        residue["is_standard_pdb"] = True
                    else:
                        residue["is_standard_pdb"] = False

            topology.atoms.append(chemfiles_atom)

        # (2) add residues to the topology
        for residue in residues.values():
            topology.residues.append(residue)

        # (3) add bonds to the topology
        for bond in getattr(obj, "bonds", []):
            topology.add_bond(bond.atoms[0].ix, bond.atoms[1].ix)

        return topology


class ChemfilesPicklable(chemfiles.Trajectory):
    """Chemfiles file object (read-only) that can be pickled.

    This class provides a file-like object (as returned by
    :class:`chemfiles.Trajectory`) that,
    unlike standard Python file objects,
    can be pickled. Only read mode is supported.

    When the file is pickled, path, mode, and format of the file handle
    are saved. On unpickling, the file is opened by path with mode,
    and saved format.
    This means that for a successful unpickle, the original file still has
    to be accessible with its filename.

    Note
    ----
    Can only be used with reading ('r') mode.
    Upon pickling, the current frame is reset. `universe.trajectory[i]` has
    to be used to return to its original frame.

    Parameters
    ----------
    filename : str
        a filename given a text or byte string.
    mode : 'r' , optional
        only 'r' can be used for pickling.
    format : '', optional
        guessed from the file extension if empty.

    Example
    -------
    ::

        f = ChemfilesPicklable(XYZ, 'r', '')
        print(f.read())
        f.close()

    can also be used as context manager::

        with ChemfilesPicklable(XYZ) as f:
            print(f.read())

    See Also
    ---------
    :class:`MDAnalysis.lib.picklable_file_io.FileIOPicklable`
    :class:`MDAnalysis.lib.picklable_file_io.BufferIOPicklable`
    :class:`MDAnalysis.lib.picklable_file_io.TextIOPicklable`
    :class:`MDAnalysis.lib.picklable_file_io.GzipPicklable`
    :class:`MDAnalysis.lib.picklable_file_io.BZ2Picklable`


    .. versionadded:: 2.0.0
    """

    def __init__(self, path, mode="r", format=""):
        if mode != "r":
            raise ValueError("Only read mode ('r') files can be pickled.")
        super().__init__(path=path, mode=mode, format=format)

    def __getstate__(self):
        return self.path, self._Trajectory__mode, self._Trajectory__format

    def __setstate__(self, args):
        self.__init__(*args)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest
import numpy as np
from numpy.testing import (
    assert_almost_equal,
)

from MDAnalysisTests.datafiles import MMTF, MMTF_gz, MMTF_skinny2

from MDAnalysis.coordinates.MMTF import MMTFReader


class TestMMTFReader(object):
    @pytest.fixture(scope="class")
    def r(self):
        return MMTFReader(MMTF)

    def test_read_frame_size(self, r):
        assert r.ts.n_atoms == 512

    def test_read_positions(self, r):
        assert_almost_equal(
            r.ts.positions[0], np.array([-0.798, 12.632, 23.231]), decimal=4
        )
        assert_almost_equal(
            r.ts.positions[-1], np.array([10.677, 15.517, 11.1]), decimal=4
        )

    def test_velocities(self, r):
        assert not r.ts.has_velocities

    def test_forces(self, r):
        assert not r.ts.has_forces

    def test_len(self, r):
        # should be single frame
        assert len(r) == 1


class TestMMTFReaderGZ(object):
    @pytest.fixture(scope="class")
    def r(self):
        return MMTFReader(MMTF_gz)

    def test_read_frame_size(self, r):
        assert r.ts.n_atoms == 1140

    def test_read_positions(self, r):
        assert_almost_equal(
            r.ts.positions[0], np.array([38.428, 16.440, 28.841]), decimal=4
        )
        assert_almost_equal(
            r.ts.positions[-1], np.array([36.684, 27.024, 20.468]), decimal=4
        )

    def test_velocities(self, r):
        assert not r.ts.has_velocities

    def test_forces(self, r):
        assert not r.ts.has_forces

    def test_len(self, r):
        # should be single frame
        assert len(r) == 1


def test_dimensionless():
    r = MMTFReader(MMTF_skinny2)

    assert r.ts.dimensions is None


def test_deprecation():
    wmsg = "The MMTF Reader is deprecated"

    with pytest.warns(DeprecationWarning, match=wmsg):
        _ = MMTFReader(MMTF_skinny2)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
DESRES file format --- :mod:`MDAnalysis.coordinates.DMS`
========================================================

Classes to read DESRES_ Molecular Structure file format (DMS_)
coordinate files (as used by the Desmond_ MD package).

.. _DESRES: http://www.deshawresearch.com
.. _Desmond: http://www.deshawresearch.com/resources_desmond.html
.. _DMS: http://www.deshawresearch.com/Desmond_Users_Guide-0.7.pdf
"""
import numpy as np
import sqlite3

from . import base
from .core import triclinic_box


class DMSReader(base.SingleFrameReaderBase):
    """
    Reads both coordinates and velocities.

    .. versionchanged:: 0.11.0
       Frames now 0-based instead of 1-based
    """

    format = "DMS"
    units = {"time": None, "length": "A", "velocity": "A/ps"}

    def get_coordinates(self, cur):
        cur.execute("SELECT * FROM particle")
        particles = cur.fetchall()
        return [(p["x"], p["y"], p["z"]) for p in particles]

    def get_particle_by_columns(self, cur, columns=None):
        if columns is None:
            columns = ["x", "y", "z"]
        cur.execute("SELECT * FROM particle")
        particles = cur.fetchall()
        return [tuple([p[c] for c in columns]) for p in particles]

    def get_global_cell(self, cur):
        cur.execute("SELECT * FROM global_cell")
        rows = cur.fetchall()
        assert len(rows) == 3
        x = [row["x"] for row in rows]
        y = [row["y"] for row in rows]
        z = [row["z"] for row in rows]
        return {"x": x, "y": y, "z": z}

    def _read_first_frame(self):
        coords_list = None
        velocities_list = None

        def dict_factory(cursor, row):
            d = {}
            for idx, col in enumerate(cursor.description):
                d[col[0]] = row[idx]
            return d

        with sqlite3.connect(self.filename) as con:
            # This will return dictionaries instead of tuples, when calling cur.fetch() or fetchall()
            con.row_factory = dict_factory
            cur = con.cursor()
            coords_list = self.get_coordinates(cur)
            velocities_list = self.get_particle_by_columns(
                cur, columns=["vx", "vy", "vz"]
            )
            unitcell = self.get_global_cell(cur)

        if not coords_list:
            raise IOError("Found no coordinates")
        self.n_atoms = len(coords_list)

        velocities = np.array(velocities_list, dtype=np.float32)
        if not velocities.any():
            velocities = None

        self.ts = self._Timestep.from_coordinates(
            np.array(coords_list, dtype=np.float32),
            velocities=velocities,
            **self._ts_kwargs,
        )
        self.ts.frame = 0  # 0-based frame number

        self.ts.dimensions = triclinic_box(
            unitcell["x"], unitcell["y"], unitcell["z"]
        )

        if self.convert_units:
            self.convert_pos_from_native(self.ts._pos)  # in-place !
            if self.ts.dimensions is not None:
                self.convert_pos_from_native(
                    self.ts.dimensions[:3]
                )  # in-place !
            if self.ts.has_velocities:
                # converts nm/ps to A/ps units
                self.convert_velocities_from_native(self.ts._velocities)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pickle

import os
import subprocess
import errno
from collections import defaultdict
from io import StringIO
import warnings

import numpy as np
from numpy.testing import (
    assert_allclose,
    assert_almost_equal,
    assert_equal,
    assert_array_equal,
)
import pytest

from MDAnalysisTests import make_Universe
from MDAnalysisTests.datafiles import (
    PSF,
    DCD,
    PSF_BAD,
    PDB_small,
    PDB_chainidrepeat,
    GRO,
    TRR,
    two_water_gro,
    two_water_gro_nonames,
    TRZ,
    TRZ_psf,
    PDB,
    MMTF,
    CONECT,
    PDB_conect,
    ITP_tip5p,
    TPR,
)

import MDAnalysis as mda
import MDAnalysis.coordinates
from MDAnalysis.topology.base import TopologyReaderBase
from MDAnalysis.transformations import translate
from MDAnalysisTests import assert_nowarns
from MDAnalysis.exceptions import NoDataError
from MDAnalysis.core.topologyattrs import AtomStringAttr
from MDAnalysisTests.util import get_userid


class IOErrorParser(TopologyReaderBase):
    def parse(self, **kwargs):
        raise IOError("Useful information")


# This string is not in the `TestUniverseCreation` class or its method because of problems
# with whitespace. Extra indentations make the string unreadable.
CHOL_GRO = """\
Single cholesterol molecule
8
  153CHOL   ROH 1793   6.558   2.936   4.005 -0.1044 -0.1252  0.0966
  153CHOL    R1 1794   6.591   2.999   4.279  0.0998  0.1297  0.0158
  153CHOL    R2 1795   6.657   2.810   4.469  0.0780  0.2817  0.1592
  153CHOL    R3 1796   6.859   2.983   4.524  0.2233  0.2112 -0.1215
  153CHOL    R4 1797   6.804   2.849   4.779  0.4156  0.3232  0.0001
  153CHOL    R5 1798   6.810   3.064   4.744  0.4811  0.3182 -0.0905
  153CHOL    C1 1799   7.080   3.034   5.012  0.7486  0.2811 -0.3559
  153CHOL    C2 1800   6.993   3.163   5.284  0.3677 -0.2104 -0.0829
10 10 10
"""


class TestUniverseCreation(object):
    # tests concerning Universe creation and errors encountered
    def test_load(self):
        # Universe(top, trj)
        u = mda.Universe(PSF, PDB_small)
        assert_equal(len(u.atoms), 3341, "Loading universe failed somehow")

    def test_load_topology_stringio(self):
        u = mda.Universe(StringIO(CHOL_GRO), format="GRO")
        assert_equal(
            len(u.atoms), 8, "Loading universe from StringIO failed somehow"
        )
        assert_equal(
            u.trajectory.ts.positions[0],
            np.array([65.580002, 29.360001, 40.050003], dtype=np.float32),
        )

    def test_load_trajectory_stringio(self):
        u = mda.Universe(
            StringIO(CHOL_GRO),
            StringIO(CHOL_GRO),
            format="GRO",
            topology_format="GRO",
        )
        assert_equal(
            len(u.atoms), 8, "Loading universe from StringIO failed somehow"
        )

    def test_make_universe_stringio_no_format(self):
        # Loading from StringIO without format arg should raise TypeError
        with pytest.raises(TypeError):
            mda.Universe(StringIO(CHOL_GRO))

    def test_Universe_no_trajectory_AE(self):
        # querying trajectory without a trajectory loaded (only topology)
        u = make_Universe()
        with pytest.raises(AttributeError):
            getattr(u, "trajectory")

    def test_Universe_topology_unrecognizedformat_VE(self):
        with pytest.raises(ValueError):
            mda.Universe("some.file.without.parser_or_coordinate_extension")

    def test_Universe_topology_unrecognizedformat_VE_msg(self):
        try:
            mda.Universe("some.file.without.parser_or_coordinate_extension")
        except ValueError as e:
            assert "isn't a valid topology format" in e.args[0]
        else:
            raise AssertionError

    def test_Universe_topology_IE(self):
        with pytest.raises(IOError):
            mda.Universe("thisfile", topology_format=IOErrorParser)

    def test_Universe_topology_IE_msg(self):
        # should get the original error, as well as Universe error
        try:
            mda.Universe("thisfile", topology_format=IOErrorParser)
        except IOError as e:
            assert "Failed to load from the topology file" in e.args[0]
            assert "Useful information" in e.args[0]
        else:
            raise AssertionError

    def test_Universe_filename_IE_msg(self):
        # check for non existent file
        try:
            mda.Universe("thisfile.xml")
        except IOError as e:
            assert_equal("No such file or directory", e.strerror)
        else:
            raise AssertionError

    def test_Universe_invalidfile_IE_msg(self, tmpdir):
        # check for invalid file (something with the wrong content)
        with tmpdir.as_cwd():
            with open("invalid.file.tpr", "w") as temp_file:
                temp_file.write("plop")
            try:
                mda.Universe("invalid.file.tpr")
            except IOError as e:
                assert "file or cannot be recognized" in e.args[0]
            else:
                raise AssertionError

    @pytest.mark.skipif(
        get_userid() == 0, reason="cannot permisssionerror as root"
    )
    def test_Universe_invalidpermissionfile_IE_msg(self, tmpdir):
        # check for file with invalid permissions (eg. no read access)
        with tmpdir.as_cwd():
            temp_file = "permission.denied.tpr"
            with open(temp_file, "w"):
                pass

            if os.name == "nt":
                subprocess.call(
                    "icacls {filename} /deny Users:RX".format(
                        filename=temp_file
                    ),
                    shell=True,
                )
            else:
                os.chmod(temp_file, 0o200)

            # Issue #3221 match by PermissionError and error number instead
            with pytest.raises(PermissionError, match=f"Errno {errno.EACCES}"):
                mda.Universe("permission.denied.tpr")

    def test_load_new_VE(self):
        u = mda.Universe.empty(0)

        with pytest.raises(TypeError):
            u.load_new("thisfile", format="soup")

    def test_load_new_memory_reader_success(self):
        u = mda.Universe(GRO)
        prot = u.select_atoms("protein")
        u2 = mda.Merge(prot)
        assert (
            u2.load_new(
                [prot.positions], format=mda.coordinates.memory.MemoryReader
            )
            is u2
        )

    def test_load_new_memory_reader_fails(self):
        def load():
            u = mda.Universe(GRO)
            prot = u.select_atoms("protein")
            u2 = mda.Merge(prot)
            u2.load_new(
                [[prot.positions]], format=mda.coordinates.memory.MemoryReader
            )

        with pytest.raises(TypeError):
            load()

    def test_universe_kwargs(self):
        u = mda.Universe(PSF, PDB_small, fake_kwarg=True)
        assert_equal(len(u.atoms), 3341, "Loading universe failed somehow")

        assert u.kwargs["fake_kwarg"]

        # initialize new universe from pieces of existing one
        u2 = mda.Universe(u.filename, u.trajectory.filename, **u.kwargs)

        assert u2.kwargs["fake_kwarg"]
        assert_equal(u.kwargs, u2.kwargs)

    def test_universe_topology_class_with_coords(self):
        u = mda.Universe(PSF, PDB_small)
        u2 = mda.Universe(u._topology, PDB_small)
        assert isinstance(u2.trajectory, type(u.trajectory))
        assert_equal(u.trajectory.n_frames, u2.trajectory.n_frames)
        assert u2._topology is u._topology

    def test_universe_empty_ROMol(self):
        Chem = pytest.importorskip("rdkit.Chem")
        mol = Chem.Mol()
        u = mda.Universe(mol, format="RDKIT")
        assert len(u.atoms) == 0


class TestUniverseFromSmiles(object):
    def setup_class(self):
        pytest.importorskip("rdkit.Chem")

    def test_default(self):
        smi = "CN1C=NC2=C1C(=O)N(C(=O)N2C)C"
        u = mda.Universe.from_smiles(smi, format="RDKIT")
        assert u.atoms.n_atoms == 24
        assert len(u.bonds.indices) == 25

    def test_from_bad_smiles(self):
        with pytest.raises(SyntaxError) as e:
            u = mda.Universe.from_smiles("J", format="RDKIT")
            assert "Error while parsing SMILES" in str(e.value)

    def test_no_Hs(self):
        smi = "CN1C=NC2=C1C(=O)N(C(=O)N2C)C"
        u = mda.Universe.from_smiles(
            smi, addHs=False, generate_coordinates=False, format="RDKIT"
        )
        assert u.atoms.n_atoms == 14
        assert len(u.bonds.indices) == 15

    def test_gencoords_without_Hs_error(self):
        with pytest.raises(ValueError) as e:
            u = mda.Universe.from_smiles(
                "CCO", addHs=False, generate_coordinates=True, format="RDKIT"
            )
            assert "requires adding hydrogens" in str(e.value)

    def test_generate_coordinates_numConfs(self):
        with pytest.raises(SyntaxError) as e:
            u = mda.Universe.from_smiles("CCO", numConfs=0, format="RDKIT")
            assert "non-zero positive integer" in str(e.value)
        with pytest.raises(SyntaxError) as e:
            u = mda.Universe.from_smiles("CCO", numConfs=2.1, format="RDKIT")
            assert "non-zero positive integer" in str(e.value)

    def test_rdkit_kwargs(self):
        # test for bad kwarg:
        # Unfortunately, exceptions from Boost cannot be passed to python,
        # we cannot `from Boost.Python import ArgumentError` and use it with
        # pytest.raises(ArgumentError), so "this is the way"
        try:
            u = mda.Universe.from_smiles("CCO", rdkit_kwargs=dict(abc=42))
        except Exception as e:
            assert "did not match C++ signature" in str(e)
        else:
            raise AssertionError(
                "RDKit should have raised an ArgumentError " "from Boost"
            )
        # good kwarg
        u1 = mda.Universe.from_smiles("C", rdkit_kwargs=dict(randomSeed=42))
        u2 = mda.Universe.from_smiles("C", rdkit_kwargs=dict(randomSeed=51))
        with pytest.raises(AssertionError) as e:
            assert_equal(
                u1.trajectory.coordinate_array, u2.trajectory.coordinate_array
            )
            assert "Mismatched elements: 15 / 15 (100%)" in str(e.value)

    def test_coordinates(self):
        # We manually create the molecule to compare coordinates
        # coordinate generation is pseudo random across different machines
        # even with the same seed, so we create a molecule to compare
        # coordinates against. See PR #4640
        from rdkit import Chem
        from rdkit.Chem import AllChem

        mol = Chem.MolFromSmiles("C", sanitize=True)
        mol = Chem.AddHs(mol)
        AllChem.EmbedMultipleConfs(mol, numConfs=2, randomSeed=42)
        expected = [c.GetPositions() for c in mol.GetConformers()]

        # now the mda way
        u = mda.Universe.from_smiles(
            "C", numConfs=2, rdkit_kwargs=dict(randomSeed=42)
        )
        assert u.trajectory.n_frames == 2
        assert_allclose(u.trajectory.coordinate_array, expected, rtol=1e-7)


class TestUniverse(object):
    # older tests, still useful
    def test_load_bad_topology(self):
        # tests that Universe builds produce the right error message
        def bad_load():
            return mda.Universe(PSF_BAD, DCD)

        with pytest.raises(ValueError):
            bad_load()

    def test_load_new(self):
        u = mda.Universe(PSF, DCD)
        u.load_new(PDB_small)
        assert_equal(len(u.trajectory), 1, "Failed to load_new(PDB)")

    def test_load_new_returns_Universe(self):
        u = mda.Universe(PSF)
        result = u.load_new(PDB_small)
        assert result is u

    def test_load_new_None_returns_Universe(self):
        u = mda.Universe(PSF)
        result = u.load_new(None)
        assert result is u

    def test_load_new_TypeError(self):
        u = mda.Universe(PSF, DCD)

        def bad_load(uni):
            return uni.load_new("filename.notarealextension")

        with pytest.raises(TypeError):
            bad_load(u)

    def test_load_structure(self):
        # Universe(struct)
        ref = mda.Universe(PSF, PDB_small)
        u = mda.Universe(PDB_small)
        assert_equal(len(u.atoms), 3341, "Loading universe failed somehow")
        assert_almost_equal(u.atoms.positions, ref.atoms.positions)

    def test_load_multiple_list(self):
        # Universe(top, [trj, trj, ...])
        ref = mda.Universe(PSF, DCD)
        u = mda.Universe(PSF, [DCD, DCD])
        assert_equal(len(u.atoms), 3341, "Loading universe failed somehow")
        assert_equal(u.trajectory.n_frames, 2 * ref.trajectory.n_frames)

    def test_load_multiple_args(self):
        # Universe(top, trj, trj, ...)
        ref = mda.Universe(PSF, DCD)
        u = mda.Universe(PSF, DCD, DCD)
        assert_equal(len(u.atoms), 3341, "Loading universe failed somehow")
        assert_equal(u.trajectory.n_frames, 2 * ref.trajectory.n_frames)

    def test_pickle(self):
        u = mda.Universe(PSF, DCD)
        s = pickle.dumps(u, protocol=pickle.HIGHEST_PROTOCOL)
        new_u = pickle.loads(s)
        assert_equal(u.atoms.names, new_u.atoms.names)

    @pytest.mark.parametrize("dtype", (int, np.float32, np.float64))
    def test_set_dimensions(self, dtype):
        u = mda.Universe(PSF, DCD)
        box = np.array([10, 11, 12, 90, 90, 90], dtype=dtype)
        u.dimensions = box
        assert_allclose(u.dimensions, box)
        assert u.dimensions.dtype == np.float32


class TestTransformations(object):
    """Tests the transformations keyword"""

    def test_callable(self):
        u = mda.Universe(PSF, DCD, transformations=translate([10, 10, 10]))
        uref = mda.Universe(PSF, DCD)
        ref = translate([10, 10, 10])(uref.trajectory.ts)
        assert_almost_equal(u.trajectory.ts.positions, ref, decimal=6)

    def test_list(self):
        workflow = [translate([10, 10, 0]), translate([0, 0, 10])]
        u = mda.Universe(PSF, DCD, transformations=workflow)
        uref = mda.Universe(PSF, DCD)
        ref = translate([10, 10, 10])(uref.trajectory.ts)
        assert_almost_equal(u.trajectory.ts.positions, ref, decimal=6)


class TestGuessTopologyAttrs(object):
    def test_automatic_type_and_mass_guessing(self):
        u = mda.Universe(PDB_small)
        assert_equal(len(u.atoms.masses), 3341)
        assert_equal(len(u.atoms.types), 3341)

    def test_no_type_and_mass_guessing(self):
        u = mda.Universe(PDB_small, to_guess=())
        assert not hasattr(u.atoms, "masses")
        assert not hasattr(u.atoms, "types")

    def test_invalid_context(self):
        u = mda.Universe(PDB_small)
        with pytest.raises(KeyError):
            u.guess_TopologyAttrs(context="trash", to_guess=["masses"])

    def test_invalid_attributes(self):
        u = mda.Universe(PDB_small)
        with pytest.raises(ValueError):
            u.guess_TopologyAttrs(to_guess=["trash"])

    def test_guess_masses_before_types(self):
        u = mda.Universe(PDB_small, to_guess=("masses", "types"))
        assert_equal(len(u.atoms.masses), 3341)
        assert_equal(len(u.atoms.types), 3341)

    def test_guessing_read_attributes(self):
        u = mda.Universe(PSF)
        old_types = u.atoms.types
        u.guess_TopologyAttrs(force_guess=["types"])
        with pytest.raises(AssertionError):
            assert_equal(old_types, u.atoms.types)


class TestGuessMasses(object):
    """Tests the Mass Guesser in topology.guessers"""

    def test_universe_loading_no_warning(self):
        assert_nowarns(UserWarning, lambda x: mda.Universe(x), GRO)


class TestGuessBonds(object):
    """Test the AtomGroup methed guess_bonds

    This needs to be done both from Universe creation (via kwarg) and AtomGroup

    It needs to:
     - work if all atoms are in vdwradii table
     - fail properly if not
     - work again if vdwradii are passed.
    """

    @pytest.fixture(scope="module")
    def vdw(self):
        return {"A": 1.4, "B": 0.5}

    def _check_universe(self, u):
        """Verify that the Universe is created correctly"""
        assert_equal(len(u.bonds), 4)
        assert_equal(len(u.angles), 2)
        assert_equal(len(u.dihedrals), 0)
        assert_equal(len(u.atoms[0].bonds), 2)
        assert_equal(len(u.atoms[1].bonds), 1)
        assert_equal(len(u.atoms[2].bonds), 1)
        assert_equal(len(u.atoms[3].bonds), 2)
        assert_equal(len(u.atoms[4].bonds), 1)
        assert_equal(len(u.atoms[5].bonds), 1)
        assert "guess_bonds" in u.kwargs

    def test_universe_guess_bonds(self):
        """Test that making a Universe with guess_bonds works"""
        u = mda.Universe(two_water_gro, guess_bonds=True)
        self._check_universe(u)
        assert u.kwargs["guess_bonds"]

    def test_universe_guess_bonds_no_vdwradii(self):
        """Make a Universe that has atoms with unknown vdwradii."""
        with pytest.raises(ValueError):
            mda.Universe(two_water_gro_nonames, guess_bonds=True)

    def test_universe_guess_bonds_with_vdwradii(self, vdw):
        """Unknown atom types, but with vdw radii here to save the day"""
        u = mda.Universe(two_water_gro_nonames, guess_bonds=True, vdwradii=vdw)
        self._check_universe(u)
        assert u.kwargs["guess_bonds"]
        assert_equal(vdw, u.kwargs["vdwradii"])

    def test_universe_guess_bonds_off(self):
        u = mda.Universe(two_water_gro_nonames, guess_bonds=False)

        for attr in ("bonds", "angles", "dihedrals"):
            assert not hasattr(u, attr)
        assert not u.kwargs["guess_bonds"]

    def test_universe_guess_bonds_arguments(self):
        """Test if 'fudge_factor', and 'lower_bound' parameters
        are being passed correctly.
        """
        u = mda.Universe(two_water_gro, guess_bonds=True)

        self._check_universe(u)
        assert u.kwargs["guess_bonds"]
        assert u.kwargs["fudge_factor"]
        assert u.kwargs["lower_bound"]

    def _check_atomgroup(self, ag, u):
        """Verify that the AtomGroup made bonds correctly,
        and that the Universe got all this info
        """
        assert_equal(len(ag.bonds), 2)
        assert_equal(len(ag.angles), 1)
        assert_equal(len(ag.dihedrals), 0)
        assert_equal(len(u.bonds), 2)
        assert_equal(len(u.angles), 1)
        assert_equal(len(u.dihedrals), 0)
        assert_equal(len(u.atoms[0].bonds), 2)
        assert_equal(len(u.atoms[1].bonds), 1)
        assert_equal(len(u.atoms[2].bonds), 1)
        assert_equal(len(u.atoms[3].bonds), 0)
        assert_equal(len(u.atoms[4].bonds), 0)
        assert_equal(len(u.atoms[5].bonds), 0)

    @pytest.mark.parametrize(
        "ff, lb, nbonds",
        [(0.55, 0.1, 2), (0.9, 1.6, 1), (0.5, 0.2, 2), (0.1, 0.1, 0)],
    )
    def test_atomgroup_guess_bonds(self, ff, lb, nbonds):
        """Test an atomgroup doing guess bonds"""
        u = mda.Universe(two_water_gro)

        ag = u.atoms[:3]
        ag.guess_bonds(fudge_factor=ff, lower_bound=lb)
        # Apply '_check_atomgroup()' only in case of default values for
        # 'fudge_factor', and 'lower_bound' arguments to avoid unnecessary
        # errors.
        if ff == 0.55 and lb == 0.1:
            self._check_atomgroup(ag, u)
        assert len(ag.bonds) == nbonds

    def test_atomgroup_guess_bonds_no_vdwradii(self):
        u = mda.Universe(two_water_gro_nonames)

        ag = u.atoms[:3]
        with pytest.raises(ValueError):
            ag.guess_bonds()

    def test_atomgroup_guess_bonds_with_vdwradii(self, vdw):
        u = mda.Universe(two_water_gro_nonames)

        ag = u.atoms[:3]
        ag.guess_bonds(vdwradii=vdw)
        self._check_atomgroup(ag, u)

    def test_guess_bonds_periodicity(self):
        u = mda.Universe(two_water_gro)

        ag = u.atoms[:3]
        ag[0].position += u.dimensions[:3] * 10
        ag.guess_bonds()

        self._check_atomgroup(ag, u)

    def guess_bonds_with_to_guess(self):
        u = mda.Universe(two_water_gro)
        has_bonds = hasattr(u.atoms, "bonds")
        u.guess_TopologyAttrs(to_guess=["bonds"])
        assert not has_bonds
        assert u.atoms.bonds

    def test_guess_read_bonds(self):
        u = mda.Universe(CONECT)
        assert len(u.bonds) == 72


class TestInMemoryUniverse(object):
    def test_reader_w_timeseries(self):
        universe = mda.Universe(PSF, DCD, in_memory=True)
        assert_equal(
            universe.trajectory.timeseries(universe.atoms).shape,
            (3341, 98, 3),
            err_msg="Unexpected shape of trajectory timeseries",
        )

    def test_reader_wo_timeseries(self):
        universe = mda.Universe(GRO, TRR, in_memory=True)
        assert_equal(
            universe.trajectory.timeseries(universe.atoms).shape,
            (47681, 10, 3),
            err_msg="Unexpected shape of trajectory timeseries",
        )

    def test_reader_w_timeseries_frame_interval(self):
        universe = mda.Universe(PSF, DCD, in_memory=True, in_memory_step=10)
        assert_equal(
            universe.trajectory.timeseries(universe.atoms).shape,
            (3341, 10, 3),
            err_msg="Unexpected shape of trajectory timeseries",
        )

    def test_reader_wo_timeseries_frame_interval(self):
        universe = mda.Universe(GRO, TRR, in_memory=True, in_memory_step=3)
        assert_equal(
            universe.trajectory.timeseries(universe.atoms).shape,
            (47681, 4, 3),
            err_msg="Unexpected shape of trajectory timeseries",
        )

    def test_existing_universe(self):
        universe = mda.Universe(PDB_small, DCD)
        universe.transfer_to_memory()
        assert_equal(
            universe.trajectory.timeseries(universe.atoms).shape,
            (3341, 98, 3),
            err_msg="Unexpected shape of trajectory timeseries",
        )

    def test_frame_interval_convention(self):
        universe1 = mda.Universe(PSF, DCD)
        array1 = universe1.trajectory.timeseries(step=10)
        universe2 = mda.Universe(PSF, DCD, in_memory=True, in_memory_step=10)
        array2 = universe2.trajectory.timeseries()
        assert_equal(
            array1, array2, err_msg="Unexpected differences between arrays."
        )

    def test_slicing_with_start_stop(self):
        universe = MDAnalysis.Universe(PDB_small, DCD)
        # Skip only the last frame
        universe.transfer_to_memory(start=10, stop=20)
        assert_equal(
            universe.trajectory.timeseries(universe.atoms).shape,
            (3341, 10, 3),
            err_msg="Unexpected shape of trajectory timeseries",
        )

    def test_slicing_without_start(self):
        universe = MDAnalysis.Universe(PDB_small, DCD)
        # Skip only the last frame
        universe.transfer_to_memory(stop=10)
        assert_equal(
            universe.trajectory.timeseries(universe.atoms).shape,
            (3341, 10, 3),
            err_msg="Unexpected shape of trajectory timeseries",
        )

    def test_slicing_without_stop(self):
        universe = MDAnalysis.Universe(PDB_small, DCD)
        # Skip only the last frame
        universe.transfer_to_memory(start=10)
        print(universe.trajectory.timeseries(universe.atoms).shape)
        assert_equal(
            universe.trajectory.timeseries(universe.atoms).shape,
            (3341, 88, 3),
            err_msg="Unexpected shape of trajectory timeseries",
        )

    def test_slicing_step_without_start_stop(self):
        universe = MDAnalysis.Universe(PDB_small, DCD)
        # Skip only the last frame
        universe.transfer_to_memory(step=2)
        print(universe.trajectory.timeseries(universe.atoms).shape)
        assert_equal(
            universe.trajectory.timeseries(universe.atoms).shape,
            (3341, 49, 3),
            err_msg="Unexpected shape of trajectory timeseries",
        )

    def test_slicing_step_with_start_stop(self):
        universe = MDAnalysis.Universe(PDB_small, DCD)
        # Skip only the last frame
        universe.transfer_to_memory(start=10, stop=30, step=2)
        print(universe.trajectory.timeseries(universe.atoms).shape)
        assert_equal(
            universe.trajectory.timeseries(universe.atoms).shape,
            (3341, 10, 3),
            err_msg="Unexpected shape of trajectory timeseries",
        )

    def test_slicing_step_dt(self):
        universe = MDAnalysis.Universe(PDB_small, DCD)
        dt = universe.trajectory.dt
        universe.transfer_to_memory(step=2)
        assert_almost_equal(
            dt * 2,
            universe.trajectory.dt,
            err_msg="Unexpected in-memory timestep: "
            + "dt not updated with step information",
        )

    def test_slicing_negative_start(self):
        universe = MDAnalysis.Universe(PDB_small, DCD)
        # Skip only the last frame
        universe.transfer_to_memory(start=-10)
        print(universe.trajectory.timeseries(universe.atoms).shape)
        assert_equal(
            universe.trajectory.timeseries(universe.atoms).shape,
            (3341, 10, 3),
            err_msg="Unexpected shape of trajectory timeseries",
        )

    def test_slicing_negative_stop(self):
        universe = MDAnalysis.Universe(PDB_small, DCD)
        # Skip only the last frame
        universe.transfer_to_memory(stop=-20)
        print(universe.trajectory.timeseries(universe.atoms).shape)
        assert_equal(
            universe.trajectory.timeseries(universe.atoms).shape,
            (3341, 78, 3),
            err_msg="Unexpected shape of trajectory timeseries",
        )

    def test_transfer_to_memory_kwargs(self):
        u = mda.Universe(PSF, DCD)
        u.transfer_to_memory(example_kwarg=True)
        assert u.trajectory._kwargs["example_kwarg"]

    def test_in_memory_kwargs(self):
        u = mda.Universe(PSF, DCD, in_memory=True, example_kwarg=True)
        assert u.trajectory._kwargs["example_kwarg"]


class TestCustomReaders(object):
    """
    Can pass a reader as kwarg on Universe creation
    """

    def test_custom_reader(self):
        # check that reader passing works
        u = mda.Universe(
            TRZ_psf, TRZ, format=MDAnalysis.coordinates.TRZ.TRZReader
        )
        assert_equal(len(u.atoms), 8184)

    def test_custom_reader_singleframe(self):
        T = MDAnalysis.topology.GROParser.GROParser
        R = MDAnalysis.coordinates.GRO.GROReader
        u = mda.Universe(
            two_water_gro, two_water_gro, topology_format=T, format=R
        )
        assert_equal(len(u.atoms), 6)

    def test_custom_reader_singleframe_2(self):
        # Same as before, but only one argument to Universe
        T = MDAnalysis.topology.GROParser.GROParser
        R = MDAnalysis.coordinates.GRO.GROReader
        u = mda.Universe(two_water_gro, topology_format=T, format=R)
        assert_equal(len(u.atoms), 6)

    def test_custom_parser(self):
        # topology reader passing works
        u = mda.Universe(
            TRZ_psf,
            TRZ,
            topology_format=MDAnalysis.topology.PSFParser.PSFParser,
        )
        assert_equal(len(u.atoms), 8184)

    def test_custom_both(self):
        # use custom for both
        u = mda.Universe(
            TRZ_psf,
            TRZ,
            format=MDAnalysis.coordinates.TRZ.TRZReader,
            topology_format=MDAnalysis.topology.PSFParser.PSFParser,
        )
        assert_equal(len(u.atoms), 8184)


class TestAddTopologyAttr(object):
    @pytest.fixture()
    def universe(self):
        return make_Universe()

    def test_add_TA_fail(self, universe):
        with pytest.raises(ValueError):
            universe.add_TopologyAttr("silly")

    def test_nodefault_fail(self, universe):
        with pytest.raises(NotImplementedError):
            universe.add_TopologyAttr("bonds")

    @pytest.mark.parametrize(
        "toadd,attrname,default",
        (
            ["charge", "charges", 0.0],
            ["charges", "charges", 0.0],
            ["name", "names", ""],
            ["names", "names", ""],
            ["type", "types", ""],
            ["types", "types", ""],
            ["element", "elements", ""],
            ["elements", "elements", ""],
            ["radius", "radii", 0.0],
            ["radii", "radii", 0.0],
            ["chainID", "chainIDs", ""],
            ["chainIDs", "chainIDs", ""],
            ["tempfactor", "tempfactors", 0.0],
            ["tempfactors", "tempfactors", 0.0],
            ["mass", "masses", 0.0],
            ["masses", "masses", 0.0],
            ["charge", "charges", 0.0],
            ["charges", "charges", 0.0],
            ["bfactor", "bfactors", 0.0],
            ["bfactors", "bfactors", 0.0],
            ["occupancy", "occupancies", 0.0],
            ["occupancies", "occupancies", 0.0],
            ["altLoc", "altLocs", ""],
            ["altLocs", "altLocs", ""],
            ["resid", "resids", 1],
            ["resids", "resids", 1],
            ["resname", "resnames", ""],
            ["resnames", "resnames", ""],
            ["resnum", "resnums", 1],
            ["resnums", "resnums", 1],
            ["icode", "icodes", ""],
            ["icodes", "icodes", ""],
            ["segid", "segids", ""],
            ["segids", "segids", ""],
        ),
    )
    def test_add_charges(self, universe, toadd, attrname, default):
        universe.add_TopologyAttr(toadd)

        assert hasattr(universe.atoms, attrname)
        assert getattr(universe.atoms, attrname)[0] == default

    @pytest.mark.parametrize(
        "attr,values",
        (
            ("bonds", [(1, 0), (1, 2)]),
            ("bonds", [[1, 0], [1, 2]]),
            ("bonds", set([(1, 0), (1, 2)])),
            ("angles", [(1, 0, 2), (1, 2, 3), (2, 1, 4)]),
            ("dihedrals", [[1, 2, 3, 1], (3, 1, 5, 2)]),
            ("impropers", [[1, 2, 3, 1], (3, 1, 5, 2)]),
        ),
    )
    def test_add_connection(self, universe, attr, values):
        universe.add_TopologyAttr(attr, values)
        assert hasattr(universe, attr)
        attrgroup = getattr(universe, attr)
        assert len(attrgroup) == len(values)
        for x in attrgroup:
            ix = x.indices
            assert ix[0] <= ix[-1]

    @pytest.mark.parametrize(
        "attr,values",
        (
            ("bonds", [(1, 0, 0), (1, 2)]),
            ("bonds", [["x", "y"], [1, 2]]),
            ("bonds", "rubbish"),
            ("bonds", [[1.01, 2.0]]),
            ("angles", [(1, 0), (1, 2)]),
            ("angles", "rubbish"),
            ("dihedrals", [[1, 1, 1, 0.1]]),
            ("impropers", [(1, 2, 3)]),
        ),
    )
    def test_add_connection_error(self, universe, attr, values):
        with pytest.raises(ValueError):
            universe.add_TopologyAttr(attr, values)

    def test_add_attr_length_error(self, universe):
        with pytest.raises(ValueError):
            universe.add_TopologyAttr(
                "masses", np.array([1, 2, 3], dtype=np.float64)
            )


class TestDelTopologyAttr(object):
    @pytest.fixture()
    def universe(self):
        u = make_Universe(("masses", "charges", "resnames"))
        u.add_TopologyAttr("bonds", [(0, 1)])
        return u

    def test_del_TA_fail(self, universe):
        with pytest.raises(ValueError, match="Unrecognised"):
            universe.del_TopologyAttr("silly")

    def test_absent_fail(self, universe):
        with pytest.raises(ValueError, match="not in Universe"):
            universe.del_TopologyAttr("angles")

    def test_wrongtype_fail(self, universe):
        with pytest.raises(ValueError, match="must be str or TopologyAttr"):
            universe.del_TopologyAttr(list)

    @pytest.mark.parametrize(
        "todel,attrname",
        [
            ("charge", "charges"),
            ("charges", "charges"),
            ("bonds", "bonds"),
        ],
    )
    def test_del_str(self, universe, todel, attrname):
        assert hasattr(universe.atoms, attrname)
        universe.del_TopologyAttr(todel)
        assert not hasattr(universe.atoms, attrname)

    def test_del_attr(self, universe):
        assert hasattr(universe.atoms, "resnames")
        assert hasattr(universe.residues, "resnames")
        assert hasattr(universe.segments, "resnames")
        assert hasattr(universe.atoms[0], "resname")
        assert hasattr(universe.residues[0], "resname")
        universe.del_TopologyAttr(universe._topology.resnames)
        assert not hasattr(universe.atoms, "resnames")
        assert not hasattr(universe.residues, "resnames")
        assert not hasattr(universe.segments, "resnames")
        assert not hasattr(universe.atoms[0], "resname")
        assert not hasattr(universe.residues[0], "resname")

    def test_del_transplants(self, universe):
        atom = universe.atoms[0]
        assert hasattr(atom, "fragindex")
        universe.del_TopologyAttr(universe.bonds)
        assert not hasattr(atom, "fragindex")
        assert not hasattr(universe.atoms, "fragindices")

    def test_del_attr_error(self, universe):
        assert not hasattr(universe._topology, "elements")
        with pytest.raises(AttributeError):
            universe._topology.del_TopologyAttr("elements")

    def test_del_attr_from_ag(self, universe):
        ag = universe.atoms[[0]]
        ag.residues.resnames = "xyz"
        universe.del_TopologyAttr("resnames")
        with pytest.raises(NoDataError):
            ag.resnames

    def test_del_func_from_universe(self, universe):
        class RootVegetable(AtomStringAttr):
            attrname = "tubers"
            singular = "tuber"
            transplants = defaultdict(list)

            def potatoes(self):
                """🥔"""
                return "potoooooooo"

            transplants["Universe"].append(("potatoes", potatoes))

        universe.add_TopologyAttr("tubers")
        assert universe.potatoes() == "potoooooooo"
        universe.del_TopologyAttr("tubers")
        with pytest.raises(AttributeError):
            universe.potatoes()


def _a_or_reversed_in_b(a, b):
    """
    Check if smaller array ``a`` or ``a[::-1]`` is in b
    """
    return (a == b).all(1).any() or (a[::-1] == b).all(1).any()


class TestAddTopologyObjects(object):

    small_atom_indices = (
        ("bonds", [[0, 1], [2, 3]]),
        ("angles", [[0, 1, 2], [3, 4, 5]]),
        ("dihedrals", [[8, 22, 1, 3], [4, 5, 6, 7], [11, 2, 3, 13]]),
        ("impropers", [[1, 6, 7, 2], [5, 3, 4, 2]]),
    )

    large_atom_indices = (
        ("bonds", [[0, 111], [22, 3]]),
        ("angles", [[0, 111, 2], [3, 44, 5]]),
        ("dihedrals", [[8, 222, 1, 3], [44, 5, 6, 7], [111, 2, 3, 13]]),
        ("impropers", [[1, 6, 771, 2], [5, 3, 433, 2]]),
    )

    @pytest.fixture()
    def empty(self):
        return make_Universe()

    @pytest.fixture()
    def universe(self):
        return mda.Universe(PSF)

    def _check_valid_added_to_empty(self, u, attr, values, to_add):
        assert not hasattr(u, attr)
        _add_func = getattr(u, "add_" + attr)
        _add_func(to_add)
        u_attr = getattr(u, attr)
        assert len(u_attr) == len(values)
        assert all(_a_or_reversed_in_b(x, u_attr.indices) for x in values)

    def _check_valid_added_to_populated(self, u, attr, values, to_add):
        assert hasattr(u, attr)
        u_attr = getattr(u, attr)
        original_length = len(u_attr)

        _add_func = getattr(u, "add_" + attr)
        _add_func(to_add)
        u_attr = getattr(u, attr)
        assert len(u_attr) == len(values) + original_length
        assert all(_a_or_reversed_in_b(x, u_attr.indices) for x in values)

    def _check_invalid_addition(self, u, attr, to_add, err_msg):
        _add_func = getattr(u, "add_" + attr)
        with pytest.raises(ValueError) as excinfo:
            _add_func(to_add)
        assert err_msg in str(excinfo.value)

    @pytest.mark.parametrize("attr,values", small_atom_indices)
    def test_add_indices_to_empty(self, empty, attr, values):
        self._check_valid_added_to_empty(empty, attr, values, values)

    def test_add_reversed_duplicates(self, empty):
        assert not hasattr(empty, "bonds")
        empty.add_bonds([[0, 1], [1, 0]])
        assert len(empty.bonds) == 1
        assert_array_equal(empty.bonds.indices, np.array([[0, 1]]))

    @pytest.mark.parametrize("attr,values", large_atom_indices)
    def test_add_indices_to_populated(self, universe, attr, values):
        self._check_valid_added_to_populated(universe, attr, values, values)

    @pytest.mark.parametrize("attr,values", small_atom_indices)
    def test_add_atomgroup_to_empty(self, empty, attr, values):
        ag = [empty.atoms[x] for x in values]
        self._check_valid_added_to_empty(empty, attr, values, ag)

    @pytest.mark.parametrize("attr,values", large_atom_indices)
    def test_add_atomgroup_to_populated(self, universe, attr, values):
        ag = [universe.atoms[x] for x in values]
        self._check_valid_added_to_populated(universe, attr, values, ag)

    @pytest.mark.parametrize("attr,values", small_atom_indices)
    def test_add_atomgroup_wrong_universe_error(
        self, universe, empty, attr, values
    ):
        ag = [empty.atoms[x] for x in values]
        self._check_invalid_addition(universe, attr, ag, "different Universes")

    @pytest.mark.parametrize("attr,values", large_atom_indices)
    def test_add_topologyobjects_to_populated(self, universe, attr, values):
        topologyobjects = [
            getattr(universe.atoms[x], attr[:-1]) for x in values
        ]
        self._check_valid_added_to_populated(
            universe, attr, values, topologyobjects
        )

    @pytest.mark.parametrize("attr,values", small_atom_indices)
    def test_add_topologyobjects_wrong_universe_error(
        self, universe, empty, attr, values
    ):
        tobj = [getattr(universe.atoms[x], attr[:-1]) for x in values]
        self._check_invalid_addition(empty, attr, tobj, "different Universes")

    @pytest.mark.parametrize("attr,values", large_atom_indices)
    def test_add_topologygroups_to_populated(self, universe, attr, values):
        topologygroup = mda.core.topologyobjects.TopologyGroup(
            np.array(values), universe
        )
        self._check_valid_added_to_populated(
            universe, attr, values, topologygroup
        )

    @pytest.mark.parametrize("attr,values", small_atom_indices)
    def test_add_topologygroup_wrong_universe_error(
        self, universe, empty, attr, values
    ):
        tg = mda.core.topologyobjects.TopologyGroup(np.array(values), universe)
        self._check_invalid_addition(empty, attr, tg, "different Universes")

    @pytest.mark.parametrize("attr,values", small_atom_indices)
    def test_add_topologygroup_different_universe(
        self, universe, empty, attr, values
    ):
        tg = mda.core.topologyobjects.TopologyGroup(np.array(values), universe)
        self._check_valid_added_to_empty(empty, attr, values, tg.to_indices())

    @pytest.mark.parametrize(
        "attr,values",
        (
            ("impropers", [[0, 111], [22, 3]]),
            ("dihedrals", [[0, 111, 2], [3, 44, 5]]),
            ("angles", [[8, 222, 1, 3], [44, 5, 6, 7], [111, 2, 3, 13]]),
            ("bonds", [[1, 6, 771, 2], [5, 3, 433, 2]]),
        ),
    )
    def test_add_wrong_topologygroup_error(self, universe, attr, values):
        arr = np.array(values)
        tg = mda.core.topologyobjects.TopologyGroup(arr, universe)
        self._check_invalid_addition(
            universe, attr, tg, "iterable of tuples with"
        )

    @pytest.mark.parametrize(
        "attr,values",
        (
            ("bonds", [[0, -111], [22, 3]]),
            ("angles", [[0, 11111, 2], [3, 44, 5]]),
            (
                "dihedrals",
                [[8, 222, 28888, 3], [44, 5, 6, 7], [111, 2, 3, 13]],
            ),
            ("impropers", [[1, 6, 77133, 2], [5, 3, 433, 2]]),
        ),
    )
    def test_add_nonexistent_indices_error(self, universe, attr, values):
        self._check_invalid_addition(
            universe, attr, values, "nonexistent atom indices"
        )

    @pytest.mark.parametrize(
        "attr,n",
        (
            ("bonds", 2),
            ("angles", 3),
            ("dihedrals", 4),
            ("impropers", 4),
        ),
    )
    def test_add_wrong_number_of_atoms_error(self, universe, attr, n):
        errmsg = (
            "{} must be an iterable of " "tuples with {} atom indices"
        ).format(attr, n)
        idx = [(0, 1), (0, 1, 2), (8, 22, 1, 3), (5, 3, 4, 2)]
        self._check_invalid_addition(universe, attr, idx, errmsg)

    def test_add_bonds_refresh_fragments(self, empty):
        with pytest.raises(NoDataError):
            getattr(empty.atoms, "fragments")

        empty.add_bonds([empty.atoms[:2]])
        assert len(empty.atoms.fragments) == len(empty.atoms) - 1

        empty.add_bonds([empty.atoms[2:4]])
        assert len(empty.atoms.fragments) == len(empty.atoms) - 2

    @pytest.mark.parametrize("attr,values", small_atom_indices)
    def test_roundtrip(self, empty, attr, values):
        _add_func = getattr(empty, "add_" + attr)
        _add_func(values)
        u_attr = getattr(empty, attr)
        assert len(u_attr) == len(values)

        _delete_func = getattr(empty, "delete_" + attr)
        _delete_func(values)
        u_attr = getattr(empty, attr)
        assert len(u_attr) == 0


class TestDeleteTopologyObjects(object):

    TOP = {
        "bonds": [(0, 1), (2, 3), (3, 4), (4, 5), (7, 8)],
        "angles": [(0, 1, 2), (3, 4, 5), (8, 2, 4)],
        "dihedrals": [
            (9, 2, 3, 4),
            (1, 3, 4, 2),
            (8, 22, 1, 3),
            (4, 5, 6, 7),
            (11, 2, 3, 13),
        ],
        "impropers": [(1, 3, 5, 2), (1, 6, 7, 2), (5, 3, 4, 2)],
    }

    existing_atom_indices = (
        ("bonds", [[0, 1], [2, 3]]),
        ("angles", [[0, 1, 2], [3, 4, 5]]),
        ("dihedrals", [[8, 22, 1, 3], [4, 5, 6, 7], [11, 2, 3, 13]]),
        ("impropers", [[1, 6, 7, 2], [5, 3, 4, 2]]),
    )
    nonexisting_atom_indices = (
        ("bonds", [[2, 3], [7, 8], [0, 4]]),
        ("angles", [[0, 1, 2], [8, 2, 8], [1, 1, 1]]),
        ("dihedrals", [[0, 0, 0, 0], [1, 1, 1, 1]]),
        (
            "impropers",
            [
                [8, 22, 1, 3],
            ],
        ),
    )

    @pytest.fixture()
    def universe(self):
        u = make_Universe(size=(125, 25, 5))
        for attr, values in self.TOP.items():
            u._add_topology_objects(attr, values)
        return u

    @pytest.fixture()
    def universe2(self):
        u = make_Universe(size=(125, 25, 5))
        for attr, values in self.TOP.items():
            u._add_topology_objects(attr, values)
        return u

    def _check_valid_deleted(self, u, attr, values, to_delete):
        u_attr = getattr(u, attr)
        original_length = len(self.TOP[attr])
        assert len(u_attr) == original_length

        _delete_func = getattr(u, "delete_" + attr)
        _delete_func(to_delete)
        u_attr = getattr(u, attr)
        assert len(u_attr) == original_length - len(values)

        not_deleted = [x for x in self.TOP[attr] if list(x) not in values]
        assert all(
            [
                x in u_attr.indices or x[::-1] in u_attr.indices
                for x in not_deleted
            ]
        )

    def _check_invalid_deleted(self, u, attr, to_delete, err_msg):
        u_attr = getattr(u, attr)
        original_length = len(self.TOP[attr])
        assert len(u_attr) == original_length
        _delete_func = getattr(u, "delete_" + attr)
        with pytest.raises(ValueError) as excinfo:
            _delete_func(to_delete)
        assert err_msg in str(excinfo.value)

    @pytest.mark.parametrize("attr,values", existing_atom_indices)
    def test_delete_valid_indices(self, universe, attr, values):
        self._check_valid_deleted(universe, attr, values, values)

    @pytest.mark.parametrize("attr,values", nonexisting_atom_indices)
    def test_delete_missing_indices(self, universe, attr, values):
        self._check_invalid_deleted(
            universe, attr, values, "Cannot delete nonexistent"
        )

    @pytest.mark.parametrize("attr,values", existing_atom_indices)
    def test_delete_valid_atomgroup(self, universe, attr, values):
        ag = [universe.atoms[x] for x in values]
        self._check_valid_deleted(universe, attr, values, ag)

    @pytest.mark.parametrize("attr,values", existing_atom_indices)
    def test_delete_atomgroup_wrong_universe_error(
        self, universe, universe2, attr, values
    ):
        ag = [universe.atoms[x] for x in values]
        self._check_invalid_deleted(universe2, attr, ag, "different Universes")

    @pytest.mark.parametrize("attr,values", nonexisting_atom_indices)
    def test_delete_missing_atomgroup(self, universe, attr, values):
        ag = [universe.atoms[x] for x in values]
        self._check_invalid_deleted(
            universe, attr, ag, "Cannot delete nonexistent"
        )

    @pytest.mark.parametrize("attr,values", existing_atom_indices)
    def test_delete_mixed_type(self, universe, attr, values):
        mixed = [universe.atoms[values[0]]] + values[1:]
        self._check_valid_deleted(universe, attr, values, mixed)

    @pytest.mark.parametrize("attr,values", existing_atom_indices)
    def test_delete_valid_topologyobjects(self, universe, attr, values):
        to = [getattr(universe.atoms[x], attr[:-1]) for x in values]
        self._check_valid_deleted(universe, attr, values, to)

    @pytest.mark.parametrize("attr,values", existing_atom_indices)
    def test_delete_topologyobjects_wrong_universe(
        self, universe, universe2, attr, values
    ):
        u1 = [getattr(universe.atoms[x], attr[:-1]) for x in values[:-1]]
        u2 = [getattr(universe2.atoms[values[-1]], attr[:-1])]
        self._check_invalid_deleted(
            universe, attr, u1 + u2, "different Universes"
        )

    @pytest.mark.parametrize("attr,values", existing_atom_indices)
    def test_delete_valid_topologygroup(self, universe, attr, values):
        arr = np.array(values)
        tg = mda.core.topologyobjects.TopologyGroup(arr, universe)
        self._check_valid_deleted(universe, attr, values, tg)

    @pytest.mark.parametrize("attr,values", existing_atom_indices)
    def test_delete_topologygroup_wrong_universe_error(
        self, universe, universe2, attr, values
    ):
        arr = np.array(values)
        tg = mda.core.topologyobjects.TopologyGroup(arr, universe2)
        self._check_invalid_deleted(universe, attr, tg, "different Universes")

    @pytest.mark.parametrize("attr,values", existing_atom_indices)
    def test_delete_topologygroup_different_universe(
        self, universe, universe2, attr, values
    ):
        arr = np.array(values)
        tg = mda.core.topologyobjects.TopologyGroup(arr, universe2)
        self._check_valid_deleted(universe, attr, values, tg.to_indices())

    @pytest.mark.parametrize(
        "attr,n",
        (
            ("bonds", 2),
            ("angles", 3),
            ("dihedrals", 4),
            ("impropers", 4),
        ),
    )
    def test_delete_wrong_number_of_atoms_error(self, universe, attr, n):
        idx = [(0, 1), (0, 1, 2), (8, 22, 1, 3), (5, 3, 4, 2)]
        errmsg = (
            "{} must be an iterable of " "tuples with {} atom indices"
        ).format(attr, n)
        self._check_invalid_deleted(universe, attr, idx, errmsg)

    @pytest.mark.parametrize("attr,values", existing_atom_indices)
    def test_delete_missing_attr(self, attr, values):
        u = make_Universe()
        assert not hasattr(u, attr)
        _delete_func = getattr(u, "delete_" + attr)
        with pytest.raises(ValueError) as excinfo:
            _delete_func(values)
        assert "There are no" in str(excinfo.value)

    def test_delete_bonds_refresh_fragments(self, universe):
        n_fragments = len(universe.atoms.fragments)
        universe.delete_bonds([universe.atoms[[2, 3]]])
        assert len(universe.atoms.fragments) == n_fragments + 1

    @pytest.mark.parametrize(
        "filename, n_bonds", [(CONECT, 72), (PDB_conect, 8)]
    )
    def test_delete_all_bonds(self, filename, n_bonds):
        u = mda.Universe(filename)
        assert len(u.bonds) == n_bonds
        u.delete_bonds(u.bonds)
        assert len(u.bonds) == 0

    @pytest.mark.parametrize("attr,values", existing_atom_indices)
    def test_roundtrip(self, universe, attr, values):
        u_attr = getattr(universe, attr)
        original_length = len(self.TOP[attr])
        assert len(u_attr) == original_length

        _delete_func = getattr(universe, "delete_" + attr)
        _delete_func(values)
        nu_attr = getattr(universe, attr)
        assert len(nu_attr) == original_length - len(values)

        _add_func = getattr(universe, "add_" + attr)
        _add_func(values)
        nu_attr = getattr(universe, attr)
        assert len(nu_attr) == original_length

        assert_array_equal(u_attr.indices, nu_attr.indices)


class TestAllCoordinatesKwarg(object):
    @pytest.fixture(scope="class")
    def u_GRO_TRR(self):
        return mda.Universe(GRO, TRR)

    @pytest.fixture(scope="class")
    def u_GRO_TRR_allcoords(self):
        return mda.Universe(GRO, TRR, all_coordinates=True)

    @pytest.fixture(scope="class")
    def u_GRO(self):
        return mda.Universe(GRO)

    def test_all_coordinates_length(self, u_GRO_TRR, u_GRO_TRR_allcoords):
        # length with all_coords should be +1
        assert len(u_GRO_TRR.trajectory) + 1 == len(
            u_GRO_TRR_allcoords.trajectory
        )

    def test_all_coordinates_frame(self, u_GRO_TRR_allcoords, u_GRO):
        # check that first frame in u(GRO, TRR, allcords)
        # are the coordinates from GRO
        assert_array_equal(
            u_GRO_TRR_allcoords.atoms.positions, u_GRO.atoms.positions
        )

    def test_second_frame(self, u_GRO_TRR_allcoords, u_GRO_TRR):
        # check that second frame in u(GRO, TRR, allcoords)
        # are the coordinates from TRR[0]
        u_GRO_TRR_allcoords.trajectory[1]

        assert_array_equal(
            u_GRO_TRR_allcoords.atoms.positions, u_GRO_TRR.atoms.positions
        )


class TestEmpty(object):
    def test_empty(self):
        u = mda.Universe.empty(10)

        assert len(u.atoms) == 10
        assert len(u.residues) == 1
        assert len(u.segments) == 1

    def test_empty_extra(self):
        u = mda.Universe.empty(
            n_atoms=12,
            n_residues=3,
            n_segments=2,
            atom_resindex=np.array([0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2]),
            residue_segindex=np.array([0, 0, 1]),
        )

        assert len(u.atoms) == 12

        assert len(u.residues) == 3
        assert len(u.residues[0].atoms) == 5
        assert len(u.residues[1].atoms) == 5
        assert len(u.residues[2].atoms) == 2

        assert len(u.segments) == 2
        assert len(u.segments[0].atoms) == 10
        assert len(u.segments[1].atoms) == 2

    def test_no_resindex_warning(self):
        with pytest.warns(UserWarning):
            u = mda.Universe.empty(n_atoms=10, n_residues=2, n_segments=1)

    def test_no_segindex_warning(self):
        res = np.array([0, 0, 0, 0, 0, 1, 1, 1, 1, 1])

        with pytest.warns(UserWarning):
            u = mda.Universe.empty(
                n_atoms=10, n_residues=2, n_segments=2, atom_resindex=res
            )

    def test_no_trivial_warning(self):
        """
        Make sure that no warning is raised about atom_resindex and
        residue_segindex when n_residues or n_segments is equal to 1.
        """
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            u = mda.Universe.empty(n_atoms=10, n_residues=1, n_segments=1)

    def test_trajectory(self):
        u = mda.Universe.empty(10, trajectory=True)

        assert len(u.atoms) == 10
        assert u.atoms.positions.shape == (10, 3)

    def test_trajectory_multiple_frames(self):
        u = mda.Universe.empty(10, n_frames=5)

        assert_equal(u.trajectory.n_frames, 5)

    def test_trajectory_iteration(self):
        u = mda.Universe.empty(10, trajectory=True)

        assert len(u.trajectory) == 1
        timesteps = []
        for ts in u.trajectory:
            timesteps.append(ts.frame)
        assert len(timesteps) == 1

    def test_velocities(self):
        u = mda.Universe.empty(10, trajectory=True, velocities=True)

        assert u.atoms.positions.shape == (10, 3)
        assert u.atoms.velocities.shape == (10, 3)

    def test_forces(self):
        u = mda.Universe.empty(10, trajectory=True, forces=True)

        assert u.atoms.positions.shape == (10, 3)
        assert u.atoms.forces.shape == (10, 3)

    def test_empty_no_atoms(self):
        u = mda.Universe.empty(0)
        assert len(u.atoms) == 0
        assert len(u.residues) == 0
        assert len(u.segments) == 0

    def test_empty_creation_raises_error(self):
        with pytest.raises(TypeError) as exc:
            u = mda.Universe()
        assert "Universe.empty" in str(exc.value)


def test_deprecate_b_tempfactors():
    u = mda.Universe(PDB)
    values = np.arange(len(u.atoms))
    with pytest.warns(DeprecationWarning, match="use the tempfactor"):
        u.add_TopologyAttr("bfactors", values)
    assert_array_equal(u.atoms.tempfactors, values)


class Thingy:
    def __init__(self, val):
        self.v = val


class ThingyParser(TopologyReaderBase):
    format = "THINGY"

    @staticmethod
    def _format_hint(thing):
        return isinstance(thing, Thingy)

    def parse(self, **kwargs):
        return mda.core.topology.Topology(n_atoms=10)


class TestOnlyTopology:
    def test_only_top(self):
        # issue 3443
        t = Thingy(20)

        with pytest.warns(UserWarning, match="No coordinate reader found for"):
            u = mda.Universe(t, to_guess=())

        assert len(u.atoms) == 10


class TestUniverseSetGroups:
    @pytest.fixture()
    def bad_seg(self):
        bad_seg_str = """\
ATOM    659  N   THR A 315      22.716  15.055  -1.000  1.00 16.08         A N
ATOM    660  CA  THR A 315      22.888  13.803  -0.302  1.00  0.00           C
ATOM    661  C   THR A 315      22.006  12.700  -0.882  1.00  0.00           C
ATOM    662  O   THR A 316      21.138  12.959  -1.727  1.00 16.25         A O
ATOM    663  CB  THR B 314      22.481  13.956   1.182  1.00  0.00         B C
ATOM    664  CG2 THR B 315      23.384  14.924   1.927  1.00  0.00           C
ATOM    665  OG1 THR B 315      21.172  14.548   1.274  1.00  0.00           O
"""
        return mda.Universe(StringIO(bad_seg_str), format="PDB")

    @pytest.fixture()
    def good(self):
        good_str = """\
ATOM    659  N   THR A 315      22.716  15.055  -1.000  1.00 16.08         A N
ATOM    660  CA  THR A 315      22.888  13.803  -0.302  1.00 15.13         A C
ATOM    661  C   THR A 315      22.006  12.700  -0.882  1.00 15.69         A C
ATOM    662  O   THR A 316      21.138  12.959  -1.727  1.00 16.25         A O
ATOM    663  CB  THR B 314      22.481  13.956   1.182  1.00 16.22         B C
ATOM    664  CG2 THR B 315      22.874  15.310   1.747  1.00 17.32         B C
ATOM    665  OG1 THR B 315      21.047  13.922   1.304  1.00 15.14         B O
"""
        return mda.Universe(StringIO(good_str), format="PDB")

    @pytest.fixture()
    def bad_gro(self):
        bad_gro_str = """\
Written by MDAnalysis
    7
    2THR      N    1   2.272   1.506  -0.100
    2THR     CA    2   2.289   1.380  -0.030
    2THR      C    3   2.201   1.270  -0.088
    3THR      O    4   2.114   1.296  -0.173
    1THR     CB    5   2.248   1.396   0.118
    2THR    CG2    6   2.287   1.531   0.175
    2THR    OG1    7   2.105   1.392   0.130
   3.00000   3.00000   3.00000
"""
        return mda.Universe(StringIO(bad_gro_str), format="GRO")

    @pytest.fixture()
    def itp(self):
        return mda.Universe(ITP_tip5p, HW2_CHARGE=2, infer_system=True)

    @pytest.fixture()
    def tpr(self):
        return mda.Universe(TPR)

    def test_do_nothing(self, good):
        # check do nothing (check the warning message)
        with pytest.warns(
            UserWarning,
            match="Not setting groups. Please provide atomwise_resids or "
            "atomwise_segids",
        ):
            good.set_groups()

            # check segments
            assert_equal(["A", "B"], good.segments.segids)
            # check residues
            assert_equal([315, 316, 314, 315], good.residues.resids)

    def test_set_segments(self, bad_seg, good):
        # check bad_seg with good
        # [BEFORE] 5 segments and 5 residues in bad_seg
        assert len(bad_seg.segments) == 5  # 5 segments
        assert_equal(["A", "", "A", "B", ""], bad_seg.segments.segids)
        assert len(bad_seg.residues) == 5  # 5 residues
        assert_equal([315, 315, 316, 314, 315], bad_seg.residues.resids)
        assert_equal(
            ["THR", "THR", "THR", "THR", "THR"], bad_seg.residues.resnames
        )
        assert_equal(["A", "", "A", "B", ""], bad_seg.residues.segids)
        assert len(bad_seg.atoms) == 7  # 7 atoms
        assert_equal([315, 315, 315, 316, 314, 315, 315], bad_seg.atoms.resids)
        assert_equal(["THR"] * 7, bad_seg.atoms.resnames)
        assert_equal(["A", "", "", "A", "B", "", ""], bad_seg.atoms.segids)
        assert_equal(
            ["N", "CA", "C", "O", "CB", "CG2", "OG1"], bad_seg.atoms.names
        )
        assert_equal(
            ["A", "A", "A", "A", "B", "B", "B"], bad_seg.atoms.chainIDs
        )
        original_attrs = bad_seg._topology.attrs

        # [ACT] set new segids
        bad_seg.set_groups(atomwise_segids=["A", "A", "A", "A", "B", "B", "B"])

        # [AFTER] 2 segments and 4 residues in bad_seg
        # segment level
        assert len(bad_seg.segments) == len(good.segments)
        assert len(bad_seg.segments) == 2  # 2 segments
        assert_equal(bad_seg.segments.segids, good.segments.segids)

        # residue level
        assert len(bad_seg.residues) == len(good.residues)
        assert len(bad_seg.residues) == 4  # 4 residues
        assert_equal(bad_seg.residues.resids, good.residues.resids)
        assert_equal(bad_seg.residues.resnames, good.residues.resnames)
        assert_equal(bad_seg.residues.segids, good.residues.segids)

        # atom level
        assert len(bad_seg.atoms) == len(good.atoms)
        assert len(bad_seg.atoms) == 7  # 7 atoms
        assert_equal(bad_seg.atoms.resids, good.atoms.resids)
        assert_equal(bad_seg.atoms.resnames, good.atoms.resnames)
        assert_equal(bad_seg.atoms.segids, good.atoms.segids)
        assert_equal(bad_seg.atoms.names, good.atoms.names)
        assert_equal(bad_seg.atoms.chainIDs, good.atoms.chainIDs)

        # check attrs
        assert len(bad_seg._topology.attrs) == len(original_attrs)
        assert_equal(
            sorted([each.attrname for each in bad_seg._topology.attrs]),
            sorted([each.attrname for each in original_attrs]),
        )

    def test_set_residues(self, bad_gro, good):
        # check bad_gro with good
        # [BEFORE] resids in bad_gro
        assert_equal([2, 3, 1, 2], bad_gro.residues.resids)
        assert_equal([2, 2, 2, 3, 1, 2, 2], bad_gro.atoms.resids)
        original_attrs = bad_gro._topology.attrs

        # [ACT]
        bad_gro.set_groups(
            atomwise_resids=[315, 315, 315, 316, 314, 315, 315],
        )

        # [AFTER] resids in bad_gro
        assert len(bad_gro.residues) == len(good.residues)
        assert_equal(bad_gro.residues.resids, good.residues.resids)
        assert len(bad_gro.atoms) == len(good.atoms)
        assert_equal(bad_gro.atoms.resids, good.atoms.resids)
        assert len(bad_gro._topology.attrs) == len(original_attrs)
        assert_equal(
            sorted([each.attrname for each in bad_gro._topology.attrs]),
            sorted([each.attrname for each in original_attrs]),
        )

    def test_set_groups_both(self, bad_gro, good):
        # test: both atomwise_segids and atomwise_resids exist
        # check bad_gro with good
        # [BEFORE] 1 segment in bad_gro
        assert len(bad_gro.segments) == 1
        assert_equal(["SYSTEM"], bad_gro.segments.segids)
        assert len(bad_gro.residues) == 4
        assert_equal(["SYSTEM"] * 4, bad_gro.residues.segids)
        assert_equal([2, 3, 1, 2], bad_gro.residues.resids)
        assert len(bad_gro.atoms) == 7
        assert_equal(["SYSTEM"] * 7, bad_gro.atoms.segids)
        assert_equal([2, 2, 2, 3, 1, 2, 2], bad_gro.atoms.resids)
        original_attrs = bad_gro._topology.attrs

        # [ACT] set new segids and resids
        bad_gro.set_groups(
            atomwise_segids=good.atoms.segids,
            atomwise_resids=good.atoms.resids,
        )

        # [AFTER] 1 segments in bad_gro
        assert len(bad_gro.segments) == len(good.segments)
        assert len(bad_gro.segments) == 2
        assert_equal(bad_gro.segments.segids, good.segments.segids)
        assert len(bad_gro.residues) == len(good.residues)
        assert len(bad_gro.residues) == 4
        assert_equal(bad_gro.residues.segids, good.residues.segids)
        assert_equal(bad_gro.residues.resids, good.residues.resids)
        assert len(bad_gro.atoms) == 7
        assert_equal(bad_gro.atoms.segids, good.atoms.segids)
        assert_equal(bad_gro.atoms.resids, good.atoms.resids)
        assert len(bad_gro._topology.attrs) == len(original_attrs)
        assert_equal(
            sorted([each.attrname for each in bad_gro._topology.attrs]),
            sorted([each.attrname for each in original_attrs]),
        )

    def test_value_error(self, good):
        # [BEFORE] atomwise_segids
        assert len(good.atoms) == 7

        # [ACT] set new segids and resids
        with pytest.raises(ValueError):
            good.set_groups(
                atomwise_segids=["A", "A"],
            )

        with pytest.raises(ValueError):
            good.set_groups(
                atomwise_resids=[30, 31],
            )

    def test_with_ITP(self, itp):
        # [BEFORE] 1 segment in itp
        assert len(itp.segments) == 1
        assert_equal(["SOL"], itp.segments.segids)
        assert len(itp.residues) == 1
        assert_equal([1], itp.residues.resids)
        assert_equal([1], itp.residues.chargegroups)
        assert len(itp.atoms) == 5
        assert_equal([1] * 5, itp.atoms.resids)
        assert_equal([1] * 5, itp.atoms.chargegroups)
        original_attrs = itp._topology.attrs

        # [ACT] set new segids and resids
        itp.set_groups(
            atomwise_segids=["A", "A", "B", "B", "B"],
            atomwise_resids=[3, 3, 3, 3, 4],
        )

        # [AFTER] 2 segments in itp
        assert len(itp.segments) == 2
        assert_equal(["A", "B"], itp.segments.segids)
        assert len(itp.residues) == 3
        assert_equal(["A", "B", "B"], itp.residues.segids)
        assert_equal([3, 3, 4], itp.residues.resids)
        assert_equal([1, 1, 1], itp.residues.chargegroups)
        assert len(itp.atoms) == 5
        assert_equal([3, 3, 3, 3, 4], itp.atoms.resids)
        assert_equal([1] * 5, itp.atoms.chargegroups)
        assert len(itp._topology.attrs) == len(original_attrs)
        assert_equal(
            sorted([each.attrname for each in itp._topology.attrs]),
            sorted([each.attrname for each in original_attrs]),
        )

    def test_with_TPR(self, tpr):
        # [BEFORE]
        assert len(tpr.segments) == 3
        assert_equal(
            ["seg_0_AKeco", "seg_1_SOL", "seg_2_NA+"], tpr.segments.segids
        )
        assert len(tpr.residues) == 11302
        assert_equal(range(1, 11303, 1), tpr.residues.resids)
        orignial_res_molnums = tpr.residues.molnums
        assert len(tpr.atoms) == 47681
        orignial_resids = tpr.atoms.resids
        orignial_atm_molnums = tpr.atoms.molnums
        original_attrs = tpr._topology.attrs

        # [ACT] set new segids and resids
        sub_dict = {"seg_0_AKeco": "A", "seg_1_SOL": "B", "seg_2_NA+": "C"}
        new_segids = [sub_dict[each] for each in tpr.atoms.segids]
        tpr.set_groups(
            atomwise_segids=new_segids,
            atomwise_resids=tpr.atoms.resids + 120,
        )

        # [AFTER]
        assert len(tpr.segments) == 3
        assert_equal(["A", "B", "C"], tpr.segments.segids)
        assert len(tpr.residues) == 11302
        assert_equal(range(1 + 120, 11303 + 120, 1), tpr.residues.resids)
        assert_equal(orignial_res_molnums, tpr.residues.molnums)
        assert len(tpr.atoms) == 47681
        assert_equal(orignial_resids + 120, tpr.atoms.resids)
        assert_equal(orignial_atm_molnums, tpr.atoms.molnums)
        assert len(tpr._topology.attrs) == len(original_attrs)
        assert_equal(
            sorted([each.attrname for each in tpr._topology.attrs]),
            sorted([each.attrname for each in original_attrs]),
        )

    def test_no_affect_custom_attrs(self, good):
        # [BEFORE-1] check custom attributes
        with pytest.raises(NoDataError):
            good.atoms.charges
        assert len(good._topology.attrs) == 18

        # [ACT-1] assign custom attributes
        good.add_TopologyAttr("charges", [0, 1, -1, 0, -1, 0, 1])

        # [AFTER-1] check custom attributes
        assert_equal([0, 1, -1, 0, -1, 0, 1], good.atoms.charges)
        assert len(good._topology.attrs) == 19

        # [BEFORE-2] check segids and resids
        assert len(good.segments) == 2
        assert_equal(["A", "B"], good.segments.segids)
        assert len(good.residues) == 4
        assert_equal([315, 316, 314, 315], good.residues.resids)
        original_attrs = good._topology.attrs

        # [ACT-2] set new segids and resids
        good.set_groups(
            atomwise_segids=["C", "C", "C", "C", "C", "C", "C"],
            atomwise_resids=[1, 1, 1, 2, 2, 2, 3],
        )

        # [AFTER-2] check segids and resids
        assert len(good.segments) == 1
        assert_equal(["C"], good.segments.segids)
        assert len(good.residues) == 3
        assert_equal([1, 2, 3], good.residues.resids)
        assert_equal([0, 1, -1, 0, -1, 0, 1], good.atoms.charges)
        assert len(good._topology.attrs) == 19
        assert_equal(
            sorted([each.attrname for each in original_attrs]),
            sorted([each.attrname for each in good._topology.attrs]),
        )
